#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
BSD 3-Clause License

Copyright (c) 2019, Max Heikenfeld
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import warnings

warnings.filterwarnings("ignore", category=UserWarning, append=True)
warnings.filterwarnings("ignore", category=RuntimeWarning, append=True)
warnings.filterwarnings("ignore", category=FutureWarning, append=True)
from cf_units import Unit
from iris import Constraint
from iris.coords import AuxCoord
from iris.cube import CubeList
from iris.util import promote_aux_coord_to_dim_coord


def load(dataset, variable, mode="auto", **kwargs):
    return loadramscube_mult(dataset, variable)


def loadramscubelist(filenames, variable_list, **kwargs):
    cubelist_out = CubeList()
    for variable in variable_list:
        cubelist_out.append(loadramscube(filenames, variable, **kwargs))
    return cubelist_out


def loadramscube(dataset, variable, **kwargs):
    variable_cube = loadramscube_mult(dataset, variable)

    return variable_cube


def loadramscube_mult(dataset, variable, constraint=None, add_coordinates=None):
    array = dataset[variable]
    variable_dimensions = array.dims
    attributes = dataset.attrs
    cube = array.to_iris()
    coord_system = make_coord_system(attributes)

    for dim, dimension in enumerate(variable_dimensions):
        if variable_dimensions[dim] == "west_east":
            west_east = make_westeast_coord(
                attributes["DX"], attributes["WEST-EAST_PATCH_END_UNSTAG"]
            )
            cube.add_dim_coord(west_east, dim)
            projection_x_coord = make_x_coord(
                attributes["DX"],
                attributes["WEST-EAST_PATCH_END_UNSTAG"],
                coord_system=coord_system,
            )
            cube.add_aux_coord(projection_x_coord, dim)
            x_coord = AuxCoord(cube.coord("west_east").points, long_name="x", units=1)
            cube.add_aux_coord(x_coord, data_dims=cube.coord_dims("west_east"))
        elif variable_dimensions[dim] == "south_north":
            south_north = make_southnorth_coord(
                attributes["DY"], attributes["SOUTH-NORTH_PATCH_END_UNSTAG"]
            )
            cube.add_dim_coord(south_north, dim)
            projection_y_coord = make_y_coord(
                attributes["DY"],
                attributes["SOUTH-NORTH_PATCH_END_UNSTAG"],
                coord_system=coord_system,
            )
            cube.add_aux_coord(projection_y_coord, dim)
            y_coord = AuxCoord(cube.coord("south_north").points, long_name="y", units=1)
            cube.add_aux_coord(y_coord, data_dims=cube.coord_dims("south_north"))

        elif variable_dimensions[dim] == "west_east_stag":
            west_east_stag = make_westeast_stag_coord(
                attributes["DX"], attributes["WEST-EAST_PATCH_END_STAG"]
            )
            cube.add_dim_coord(west_east_stag, dim)
            projection_x_stag_coord = make_x_stag_coord(
                attributes["DX"],
                attributes["WEST-EAST_PATCH_END_STAG"],
                coord_system=coord_system,
            )
            cube.add_aux_coord(projection_x_stag_coord, dim)
            x_coord = AuxCoord(
                cube.coord("west_east_stag").points, long_name="x", units=1
            )
            cube.add_aux_coord(x_coord, data_dims=cube.coord_dims("west_east_stag"))

        elif variable_dimensions[dim] == "south_north_stag":
            south_north_stag = make_southnorth_stag_coord(
                attributes["DY"], attributes["SOUTH-NORTH_PATCH_END_STAG"]
            )
            cube.add_dim_coord(south_north_stag, dim)
            projection_y_stag_coord = make_y_stag_coord(
                attributes["DY"],
                attributes["SOUTH-NORTH_PATCH_END_STAG"],
                coord_system=coord_system,
            )
            cube.add_aux_coord(projection_y_stag_coord, dim)
            y_coord = AuxCoord(
                cube.coord("south_north_stag").points, long_name="y", units=1
            )
            cube.add_aux_coord(y_coord, data_dims=cube.coord_dims("south_north_stag"))

        elif variable_dimensions[dim] == "bottom_top":
            bottom_top = make_bottom_top_coordinate(
                attributes["BOTTOM-TOP_PATCH_END_UNSTAG"]
            )
            cube.add_dim_coord(bottom_top, dim)
            model_level_number = make_model_level_number_coordinate(
                attributes["BOTTOM-TOP_PATCH_END_UNSTAG"]
            )
            cube.add_aux_coord(model_level_number, dim)

        elif variable_dimensions[dim] == "bottom_top_stag":
            bottom_top_stag = make_bottom_top_stag_coordinate(
                attributes["BOTTOM-TOP_PATCH_END_STAG"]
            )
            cube.add_dim_coord(bottom_top_stag, dim)
            model_level_number = make_model_level_number_coordinate(
                attributes["BOTTOM-TOP_PATCH_END_STAG"]
            )
            cube.add_aux_coord(model_level_number, dim)

    if "Times" in [coord.name() for coord in cube.coords()]:
        cube.coord("Times").rename("time")
        promote_aux_coord_to_dim_coord(cube, "time")
        cube.coord("time").attributes = {}
        # only used to ensure that the time dimension has units
        cube.coord("time").units = Unit(dataset.date)

    # change latitude and longitude coordinates to 2D fields (fine andm ore consistent with other models for all static rams Simulations)
    if "GLAT" in [coord.name() for coord in cube.coords()]:
        latitude_coord = cube.extract(
            Constraint(
                time=cube.coord("time").units.num2date(cube.coord("time").points[0])
            )
        ).coord("GLAT")
        latitude_coord.rename("latitude")
        xlat_dims = list(cube.coord_dims("GLAT"))
        time_dim = cube.coord_dims("time")[0]
        xlat_dims.remove(time_dim)
        data_dims = tuple(xlat_dims)
        cube.add_aux_coord(latitude_coord, data_dims=data_dims)
        cube.remove_coord("GLAT")
    if "GLON" in [coord.name() for coord in cube.coords()]:
        longitude_coord = cube[0].coord("GLON")
        longitude_coord.rename("longitude")
        xlong_dims = list(cube.coord_dims("GLON"))
        time_dim = cube.coord_dims("time")[0]
        xlong_dims.remove(time_dim)
        data_dims = tuple(xlong_dims)
        cube.add_aux_coord(longitude_coord, data_dims=data_dims)
        cube.remove_coord("GLON")

    # bring time axis into range that can be understood by pandas (e.g. idealised simulation starting in year 0)
    # reference set to 2000-01-01_00:00:00 instead
    # date_0=cube.coord('time').units.num2date(cube.coord('time').points[0])
    # if (date_0 < datetime(1678,1,1) or date_0> datetime(2262,12,31)):
    #    cube.coord('time').units=Unit(cube.coord('time').units.name.split(' since ')[0]+' since 2000-01-01')

    cube = cube.extract(constraint)

    if add_coordinates != None:
        cube = add_aux_coordinates_multidim(
            dataset, cube, constraint=None, add_coordinates=add_coordinates
        )
    return cube


def deriveramscubelist(filenames, variable_list, **kwargs):
    cubelist_out = CubeList()
    for variable in variable_list:
        cubelist_out.append(deriveramscube(filenames, variable))
    return cubelist_out


#
# def deriveramscube(filenames,variable,**kwargs):
#    if type(filenames) is list:
#        variable_cube=deriveramscube_mult(filenames,variable,**kwargs)
#    elif type(filenames) is str:
#        variable_cube=deriveramscube_single(filenames,variable,**kwargs)
#    else:
#        raise ValueError('Type of input unknown: Must be str of list')
#    return variable_cube
#
# def deriveramscube_mult(filenames,variable,**kwargs):
#
#    cube_list=[]
#    for i in range(len(filenames)):
#        cube_list.append(deriveramscube_single(filenames[i],variable,**kwargs) )
#    for member in cube_list:
#        member.attributes={}
#    variable_cubes=CubeList(cube_list)
#    variable_cube=variable_cubes.concatenate_cube()
#    #variable_cube=variable_cube.extract(**kwargs.pop('constraint'))
#
#    return variable_cube

variable_dict_pseudonym = {}
variable_dict_pseudonym["radar_relfectivity"] = "REFL10CM"


import warnings

warnings.filterwarnings("ignore", category=UserWarning, append=True)


variable_list_derive = [
    "potential_temperature",
    "temperature",
    "air_temperature",
    "density",
    "air_density",
    "LWC",
    "IWC",
    "LWP",
    "IWP",
    "IWV",
    "airmass",
    "airmass_path",
    "layer_height",
    "volume",
    "area",
    "geopotential_height",
    "pressure",
    "air_pressure",
    "relative_humidity",
    "w_unstaggered",
    "u_unstaggered",
    "v_unstaggered",
    "maximum_reflectivity",
    "surface_precipitation",
    "surface_precipitation_average",
    "surface_precipitation_accumulated",
    "surface_precipitation_instantaneous",
    "QNCLOUD",
    "QNRAIN",
    "QNGRAUPEL",
    "QNICE",
    "QNSNOW",
]


# def deriveramscube_single(filenames,variable,**kwargs):
def deriveramscube(filenames, variable, **kwargs):
    if variable == "potential_temperature":
        variable_cube = calculate_rams_potential_temperature(filenames, **kwargs)
        # variable_cube_out=addcoordinates(filenames, 'T',variable_cube,add_coordinates)
    elif variable in ["temperature", "air_temperature"]:
        variable_cube = calculate_rams_temperature(filenames, **kwargs)
        # variable_cube_out=addcoordinates(filenames, 'T',variable_cube,add_coordinates)
    elif variable in ["density", "air_density"]:
        variable_cube = calculate_rams_density(filenames, **kwargs)
        # variable_cube_out=addcoordinates(filenames, 'T',variable_cube,add_coordinates)
    elif variable in ["pressure", "air_pressure"]:
        variable_cube = calculate_rams_pressure(filenames, **kwargs)
    elif variable == "LWC":
        variable_cube = calculate_rams_LWC(filenames, **kwargs)
        # variable_cube=addcoordinates(filenames, 'QCLOUD',variable_cube,add_coordinates)
    elif variable == "IWC":
        variable_cube = calculate_rams_IWC(filenames, **kwargs)
        # variable_cube=addcoordinates(filenames, 'QICE',variable_cube,add_coordinates)
    elif variable == "LWP":
        variable_cube = calculate_rams_LWP(filenames, **kwargs)
        # variable_cube=addcoordinates(filenames, 'OLR',variable_cube,add_coordinates)
    elif variable == "IWP":
        variable_cube = calculate_rams_IWP(filenames, **kwargs)
        # variable_cube=addcoordinates(filenames, 'OLR',variable_cube,add_coordinates)
    elif variable == "IWV":
        variable_cube = calculate_rams_IWV(filenames, **kwargs)
        # variable_cube=addcoordinates(filenames, 'OLR',variable_cube,add_coordinates)
    elif variable == "airmass":
        variable_cube = calculate_rams_airmass(filenames, **kwargs)
    elif variable == "airmass_path":
        variable_cube = calculate_rams_airmass_path(filenames, **kwargs)
    elif variable == "layer_height":
        variable_cube = calculate_rams_layerheight(filenames, **kwargs)
    elif variable == "area":
        variable_cube = calculate_rams_area(filenames, **kwargs)
    elif variable == "volume":
        variable_cube = calculate_rams_volume(filenames, **kwargs)
    elif variable == "geopotential_height":
        variable_cube = calculate_rams_geopotential_height(filenames, **kwargs)
        replace_cube = loadramscube(filenames, "T", **kwargs)
        variable_cube = replacecoordinates(variable_cube, replace_cube)

    elif variable == "geopotential_height_stag":
        variable_cube = calculate_rams_geopotential_height_stag(filenames, **kwargs)

    #    elif variable == 'geopotential_height_xstag':
    #        variable_cube=calculate_rams_geopotential_height_xstag(filenames,**kwargs)
    #        replace_cube=loadramscube(filenames,'U',**kwargs)
    #        variable_cube=replacecoordinates(variable_cube,replace_cube)
    #
    #    elif variable == 'geopotential_height_ystag':
    #        variable_cube=calculate_rams_geopotential_height_ystag(filenames,**kwargs)
    #        replace_cube=loadramscube(filenames,'V',**kwargs)
    #        variable_cube=replacecoordinates(variable_cube,replace_cube)

    elif variable == "geopotential":
        variable_cube = calculate_rams_geopotential(filenames, **kwargs)

    elif variable == "pressure_xstag":
        variable_cube = calculate_rams_pressure(filenames, **kwargs)
        replace_cube = loadramscube(filenames, "U", **kwargs)
        variable_cube = replacecoordinates(variable_cube, replace_cube)

    elif variable == "pressure_ystag":
        variable_cube = calculate_rams_pressure(filenames, **kwargs)
        replace_cube = loadramscube(filenames, "V", **kwargs)
        variable_cube = replacecoordinates(variable_cube, replace_cube)

    elif variable == "relative_humidity":
        variable_cube = calculate_rams_relativehumidity(filenames, **kwargs)
        # variable_cube_out=addcoordinates(filenames, 'T',variable_cube,add_coordinates)

    elif variable == "w_unstaggered":
        variable_cube = calculate_rams_w_unstaggered(filenames, **kwargs)
        replace_cube = loadramscube(filenames, "T", **kwargs)
        variable_cube = replacecoordinates(variable_cube, replace_cube)

    elif variable == "u_unstaggered":
        constraint = kwargs.pop("constraint", None)
        variable_cube = calculate_rams_u_unstaggered(filenames, **kwargs)
        replace_cube = loadramscube(filenames, "T", **kwargs)
        variable_cube = replacecoordinates(variable_cube, replace_cube)
        variable_cube = variable_cube.extract(constraint)

    elif variable == "v_unstaggered":
        constraint = kwargs.pop("constraint", None)
        variable_cube = calculate_rams_v_unstaggered(filenames, **kwargs)
        replace_cube = loadramscube(filenames, "T", **kwargs)
        variable_cube = replacecoordinates(variable_cube, replace_cube)
        variable_cube = variable_cube.extract(constraint)

    elif variable == "surface_precipitation_average":
        variable_cube = calculate_rams_surface_precipitation_average(
            filenames, **kwargs
        )

    elif variable == "surface_precipitation_accumulated":
        variable_cube = calculate_rams_surface_precipitation_accumulated(
            filenames, **kwargs
        )

    elif (variable == "surface_precipitation_instantaneous") or (
        variable == "surface_precipitation"
    ):
        variable_cube = calculate_rams_surface_precipitation_instantaneous(
            filenames, **kwargs
        )

        # variable_cube_out=addcoordinates(filenames, 'T',variable_cube,add_coordinates)
    elif variable == "maximum_reflectivity":
        variable_cube = calculate_rams_maximum_reflectivity(filenames, **kwargs)
    elif variable in ["QNCLOUD", "QNRAIN", "QNGRAUPEL", "QNICE", "QNSNOW"]:
        variable_cube = loadramscube(filenames, variable, **kwargs)
        variable_cube.units = "kg-1"
    else:
        raise NameError(variable, "is not a known variable")
    return variable_cube


def calculate_rams_surface_precipitation_average(filenames, **kwargs):
    from dask.array import concatenate

    surface_precip_accum = calculate_rams_surface_precipitation_accumulated(
        filenames, **kwargs
    )
    # caclulate timestep in hours
    time_coord = surface_precip_accum.coord("time")
    dt = (
        time_coord.units.num2date(time_coord.points[1])
        - time_coord.units.num2date(time_coord.points[0])
    ).total_seconds() / 3600.0
    # divide difference in precip between timesteps (in mm/h) by timestep (in h):
    surface_precip = surface_precip_accum
    surface_precip.data = (
        concatenate(
            (
                0 * surface_precip.core_data()[[0], :, :],
                surface_precip.core_data()[1:, :, :]
                - surface_precip.core_data()[:-1:, :, :],
            ),
            axis=0,
        )
        / dt
    )
    surface_precip.rename("surface_precipitation_average")
    surface_precip.units = "mm h-1"
    return surface_precip


def calculate_rams_surface_precipitation_accumulated(filenames, **kwargs):
    surface_precip_accum = loadramscube(filenames, "RAINNC", **kwargs)
    surface_precip_accum.rename("surface_precipitation_accumulated")
    return surface_precip_accum


def calculate_rams_surface_precipitation_instantaneous(filenames, **kwargs):
    from xarray import open_mfdataset

    dataset = open_mfdataset(filenames, coords="all")
    dt = dataset.attrs["DT"]
    RAINNCV = loadramscube(filenames, "RAINNCV", **kwargs)
    surface_precip = RAINNCV / dt
    surface_precip.units = "kg m-2 s-1"
    surface_precip.rename("surface_precipitation_instantaneous")
    return surface_precip


def variable_list(filenames):
    from netCDF4 import Dataset

    if type(filenames) == list:
        filenames = filenames[0]
    variable_list = list(Dataset(filenames).variables)
    return variable_list


def calculate_rams_potential_temperature(filenames, **kwargs):
    from iris import coords

    T = loadramscube(filenames, "T", **kwargs)
    T0 = coords.AuxCoord(300.0, long_name="reference_temperature", units="K")
    theta = T + T0
    theta.rename("potential_temperature")
    return theta


def calculate_rams_temperature(filenames, **kwargs):
    from iris import coords

    theta = deriveramscube(filenames, "potential_temperature", **kwargs)
    p = deriveramscube(filenames, "pressure", **kwargs)
    p0 = coords.AuxCoord(1000.0, long_name="reference_pressure", units="hPa")
    p0.convert_units(p.units)
    p1 = p / p0
    exp = 287.05 / 1005.0
    T = theta * (p1**exp)
    T.rename("air temperature")
    return T


def calculate_rams_relativehumidity(filenames, **kwargs):
    from iris import cube

    QVAPOR = loadramscube(filenames, "QVAPOR", **kwargs)
    T = deriveramscube(filenames, "temperature", **kwargs)
    p = deriveramscube(filenames, "pressure", **kwargs)
    p.convert_units("Pa")
    rh = calculate_RH(QVAPOR.core_data(), T.core_data(), p.core_data())
    RH = cube.Cube(rh, units="percent", long_name="realtive humidity")
    return RH


def calculate_RH(QVAPOR, T, p):
    from numpy import exp, maximum, minimum

    ES = 1e2 * 6.1094 * exp(17.625 * (T - 273.15) / (T - 273.15 + 243.04))
    QVS = 0.622 * ES / (p - (1.0 - 0.622) * ES)
    RH = 100 * maximum(minimum(QVAPOR / QVS, 1.0), 0.0)
    return RH


def calculate_rams_LWC(filenames, **kwargs):
    microphysics_scheme = kwargs.pop("microphysics_scheme", None)
    list_variables = ["QCLOUD", "QRAIN"]
    LWC = load_sum(filenames, list_variables, **kwargs)
    LWC.rename("liquid water content")
    # LWC.rename('mass_concentration_of_liquid_water_in_air')
    return LWC


#
def calculate_rams_IWC(filenames, **kwargs):
    microphysics_scheme = kwargs.pop("microphysics_scheme", None)
    if microphysics_scheme in [None, "morrison", "thompson"]:
        list_variables = ["QICE", "QSNOW", "QGRAUP"]
    elif microphysics_scheme in ["SBM_fast"]:
        list_variables = ["QICE", "QSNOW", "QGRAUP"]
    elif microphysics_scheme in ["SBM_full"]:
        list_variables = ["QICEC", "QICED", "QICEP", "QSNOW", "QGRAUP", "QHAIL"]
    IWC = load_sum(filenames, list_variables, **kwargs)
    IWC.rename("ice water content")
    # IWC.rename('mass_concentration_of_ice_water_in_air')

    return IWC


def calculate_rams_airmass_path(filenames, **kwargs):
    rho = deriveramscube(filenames, "density", **kwargs)
    layer_height = deriveramscube(filenames, "layer_height", **kwargs)
    Airmass = rho * layer_height
    Airmass.rename("airmass_path")
    return Airmass


def calculate_rams_airmass(filenames, **kwargs):
    rho = deriveramscube(filenames, "density", **kwargs)
    volume = deriveramscube(filenames, "volume", **kwargs)
    Airmass = rho * volume
    Airmass.rename("mass_of_air")
    return Airmass


def calculate_rams_volume(filenames, **kwargs):
    layer_height = deriveramscube(filenames, "layer_height", **kwargs)
    layer_height.add_aux_coord(
        AuxCoord(
            layer_height.coord("projection_x_coordinate").bounds[..., 1]
            - layer_height.coord("projection_x_coordinate").bounds[..., 0],
            long_name="x_diff",
            units=layer_height.coord("projection_y_coordinate").units,
        ),
        data_dims=layer_height.coord_dims("projection_x_coordinate"),
    )
    layer_height.add_aux_coord(
        AuxCoord(
            layer_height.coord("projection_y_coordinate").bounds[..., 1]
            - layer_height.coord("projection_y_coordinate").bounds[..., 0],
            long_name="y_diff",
            units=layer_height.coord("projection_y_coordinate").units,
        ),
        data_dims=layer_height.coord_dims("projection_y_coordinate"),
    )
    volume = layer_height * layer_height.coord("x_diff") * layer_height.coord("y_diff")
    volume.remove_coord("x_diff")
    volume.remove_coord("y_diff")
    volume.rename("cell_volume")
    return volume


def calculate_rams_area(filenames, **kwargs):
    from numpy import diff

    dummy = loadramscube(filenames, "OLR", **kwargs)
    dummy.data[:] = 1
    dummy.units = "1"
    dummy.add_aux_coord(
        AuxCoord(
            diff(dummy.coord("projection_x_coordinate").bounds).flatten(),
            long_name="x_diff",
            units=dummy.coord("projection_y_coordinate").units,
        ),
        data_dims=dummy.coord_dims("projection_x_coordinate"),
    )
    dummy.add_aux_coord(
        AuxCoord(
            diff(dummy.coord("projection_y_coordinate").bounds).flatten(),
            long_name="y_diff",
            units=dummy.coord("projection_y_coordinate").units,
        ),
        data_dims=dummy.coord_dims("projection_y_coordinate"),
    )
    area = dummy * dummy.coord("x_diff") * dummy.coord("y_diff")
    area.rename("cell_area")
    return area


def calculate_rams_layerheight(filenames, **kwargs):
    zH = deriveramscube(filenames, "geopotential_height_stag", **kwargs)
    bottom_top_stag = zH.coord("bottom_top_stag").points
    layer_height = (
        zH.extract(Constraint(bottom_top_stag=bottom_top_stag[1:]))
        - zH.extract(Constraint(bottom_top_stag=bottom_top_stag[:-1])).core_data()
    )
    layer_height.rename("layer_height")
    replace_cube = loadramscube(filenames, "T", **kwargs)
    layer_height = replacecoordinates(layer_height, replace_cube)
    return layer_height


def calculate_rams_LWP(filenames, **kwargs):
    from iris.analysis import SUM

    LWC = deriveramscube(filenames, "LWC", **kwargs)
    microphysics_scheme = kwargs.pop("microphysics_scheme", None)
    Airmass = deriveramscube(filenames, "airmass_path", **kwargs)
    LWP = (LWC * Airmass).collapsed(("model_level_number"), SUM)
    LWP.rename("liquid water path")
    # LWP.rename('atmosphere_mass_content_of_cloud_liquid_water')
    return LWP


def calculate_rams_IWP(filenames, **kwargs):
    from iris.analysis import SUM

    IWC = deriveramscube(filenames, "IWC", **kwargs)
    microphysics_scheme = kwargs.pop("microphysics_scheme", None)
    Airmass = deriveramscube(filenames, "airmass_path", **kwargs)
    IWP = (IWC * Airmass).collapsed(("model_level_number"), SUM)
    IWP.rename("ice water path")
    # IWP.rename('atmosphere_mass_content_of_cloud_ice_water')
    return IWP


def calculate_rams_IWV(filenames, **kwargs):
    from iris.analysis import SUM

    QVAPOR = loadramscube(filenames, "QVAPOR", **kwargs)
    Airmass = deriveramscube(filenames, "airmass_path", **kwargs)
    IWV = (QVAPOR * Airmass).collapsed(("model_level_number"), SUM)
    IWV.rename("integrated water vapour")
    # IWV.rename('atmosphere_mass_content_of_water_vapor')
    return IWV


def integrate_cube(variable, Airmass_or_dz, name=None):
    from iris.analysis import SUM

    if name is None:
        name = "integrated_" + variable.name()
    variable_integrated = variable * Airmass_or_dz
    if "geopotential_height" in [
        coord.name() for coord in variable_integrated.coords()
    ]:
        variable_integrated.remove_coord("geopotential_height")
    variable_integrated = variable_integrated.collapsed(("model_level_number"), SUM)
    variable_integrated.rename(name)
    return variable_integrated


def calculate_rams_LWP_fromcubes(LWC, Airmass):
    from iris.analysis import SUM

    LW = LWC * Airmass
    # LW.remove_coord('geopotential_height')
    # !LWP=LW.collapsed(('model_level_number'),SUM)
    LWP = collapse_removecoord(LW, "model_level_number", SUM)
    LWP.rename("liquid water path")
    # LWP.rename('atmosphere_mass_content_of_cloud_liquid_water')
    return LWP


def calculate_rams_IWP_fromcubes(IWC, Airmass):
    from iris.analysis import SUM

    IW = IWC * Airmass
    # IW.remove_coord('geopotential_height')
    # IWP=IW.collapsed(('model_level_number'),SUM)
    IWP = collapse_removecoord(IW, "model_level_number", SUM)

    IWP.rename("ice water path")
    return IWP


def calculate_rams_IWV_fromcubes(QVAPOR, Airmass):
    from iris.analysis import SUM

    VAPOR = QVAPOR * Airmass
    # VAPOR.remove_coord('geopotential_height')
    # IWV=VAPOR.collapsed(('model_level_number'),SUM)
    IWV = collapse_removecoord(VAPOR, "model_level_number", SUM)
    IWV.rename("integrated water vapor")
    return IWV


def calculate_rams_maximum_reflectivity(filenames, **kwargs):
    from iris.analysis import MAX

    REFL_10CM = loadramscube(filenames, "REFL_10CM", **kwargs)
    MAX_REFL_10CM = REFL_10CM.collapsed("model_level_number", MAX)
    MAX_REFL_10CM.rename("maximum reflectivity")
    return MAX_REFL_10CM


def calculate_rams_w_unstaggered(filenames, **kwargs):
    from iris import Constraint, cube

    w = loadramscube(filenames, "W", **kwargs)
    constraint_1 = Constraint(
        bottom_top_stag=lambda cell: cell > w.coord("bottom_top_stag").points[0]
    )
    constraint_2 = Constraint(
        bottom_top_stag=lambda cell: cell < w.coord("bottom_top_stag").points[-1]
    )
    w_unstaggered = cube.Cube(
        0.5
        * (w.extract(constraint_1).core_data() + w.extract(constraint_2).core_data()),
        standard_name="upward_air_velocity",
        units="m/s",
    )
    return w_unstaggered


def calculate_rams_u_unstaggered(filenames, **kwargs):
    from iris import Constraint, cube

    u = loadramscube(filenames, "U", **kwargs)
    constraint_1 = Constraint(
        west_east_stag=lambda cell: cell > u.coord("west_east_stag").points[0]
    )
    constraint_2 = Constraint(
        west_east_stag=lambda cell: cell < u.coord("west_east_stag").points[-1]
    )
    u_unstaggered = cube.Cube(
        0.5
        * (u.extract(constraint_1).core_data() + u.extract(constraint_2).core_data()),
        standard_name="x_wind",
        units="m/s",
    )
    return u_unstaggered


def calculate_rams_v_unstaggered(filenames, **kwargs):
    from iris import Constraint, cube

    v = loadramscube(filenames, "V", **kwargs)
    constraint_1 = Constraint(
        south_north_stag=lambda cell: cell > v.coord("south_north_stag").points[0]
    )
    constraint_2 = Constraint(
        south_north_stag=lambda cell: cell < v.coord("south_north_stag").points[-1]
    )
    v_unstaggered = cube.Cube(
        0.5
        * (v.extract(constraint_1).core_data() + v.extract(constraint_2).core_data()),
        standard_name="y_wind",
        units="m/s",
    )
    return v_unstaggered


def calculate_rams_density(filenames, **kwargs):
    from iris import coords

    if "ALT" in variable_list(filenames):
        alt = loadramscube(filenames, "ALT", **kwargs)
        rho = alt ** (-1)
    else:
        R = coords.AuxCoord(
            287.058, long_name="Specific gas constant for air", units="Joule kg^-1 K^-1"
        )
        p = deriveramscube(filenames, "pressure", **kwargs)
        T = deriveramscube(filenames, "temperature", **kwargs)
        rho = p * ((R * T) ** -1)
        rho.rename("air_density")
    return rho


def calculate_rams_pressure(filenames, **kwargs):
    P = loadramscube(filenames, "P", **kwargs)
    PB = loadramscube(filenames, "PB", **kwargs)
    p = P + PB
    p.rename("pressure")
    return p


def calculate_rams_pressure_stag(filenames, **kwargs):
    p = deriveramscube(filenames, "pressure", **kwargs)
    bottom_top = p.coord("bottom_top").points
    p_stag = 0.5 * (
        p.extract(Constraint(bottom_top=bottom_top[:-1]))
        + p.extract(Constraint(bottom_top=bottom_top[1:])).core_data()
    )
    return p_stag


def calculate_rams_pressure_xstag(filenames, **kwargs):
    p = deriveramscube(filenames, "pressure", **kwargs)
    west_east = p.coord("west_east").points
    p_xstag = 0.5 * (
        p.extract(Constraint(west_east=west_east[:-1]))
        + p.extract(Constraint(west_east=west_east[1:])).core_data()
    )
    p_xstag.rename("pressure")
    return p


def calculate_rams_pressure_ystag(filenames, **kwargs):
    p = deriveramscube(filenames, "pressure", **kwargs)
    south_north = p.coord("south_north").points
    p_ystag = 0.5 * (
        p.extract(Constraint(south_north=south_north[:-1]))
        + p.extract(Constraint(south_north=south_north[1:])).core_data()
    )
    p_ystag.rename("pressure")
    return p_ystag


def calculate_rams_geopotential(filenames, **kwargs):
    PH = loadramscube(filenames, "PH", **kwargs)
    PHB = loadramscube(filenames, "PHB", **kwargs)
    pH = PH + PHB
    pH.rename("geopotential")
    return pH


def calculate_rams_geopotential_height_stag(filenames, **kwargs):
    from iris import coords

    pH = deriveramscube(filenames, "geopotential", **kwargs)
    g = coords.AuxCoord(9.81, long_name="acceleration", units="m s^-2")
    zH = pH / g
    zH.rename("geopotential_height")
    return zH


def calculate_rams_geopotential_height(filenames, **kwargs):
    zH = deriveramscube(filenames, "geopotential_height_stag", **kwargs)
    bottom_top_stag = zH.coord("bottom_top_stag").points
    z = 0.5 * (
        zH.extract(Constraint(bottom_top_stag=bottom_top_stag[:-1]))
        + zH.extract(Constraint(bottom_top_stag=bottom_top_stag[1:])).core_data()
    )
    z.rename("geopotential_height")
    return z


def calculate_rams_geopotential_height_ystag(filenames, **kwargs):
    z = calculate_rams_geopotential_height(filenames, **kwargs)
    z_ystag = cube_interp_extendby1(z, "south_north")
    return z_ystag


def calculate_rams_geopotential_height_xstag(filenames, **kwargs):
    z = calculate_rams_geopotential_height(filenames, **kwargs)
    z_xstag = cube_interp_extendby1(z, "west_east")
    return z_xstag


def unstagger(cube_in, coord, filenames, **kwargs):
    cube_out = cube_interp_reduceby1(cube_in, coord)
    replace_cube = loadramscube(filenames, "T", **kwargs)
    cube_out = replacecoordinates(cube_out, replace_cube)


def array_interp_extendby1(array, dim):
    import numpy as np

    idx1 = [slice(None)] * (array.ndim)
    idx2 = [slice(None)] * (array.ndim)
    idx_start = [slice(None)] * (array.ndim)
    idx_end = [slice(None)] * (array.ndim)
    idx1[dim] = slice(1, None)
    idx2[dim] = slice(0, -1)
    idx_start[dim] = slice(1)
    idx_end[dim] = slice(-2, -1)
    array_out = np.concatenate(
        (array[idx_start], 0.5 * (array[idx1] + array[idx2]), array[idx_end]), axis=dim
    )
    return array_out


def array_interp_reduceby1(array, dim):
    idx = [slice(None)] * (array.ndim)
    idx1 = idx
    idx2 = idx
    idx1[dim] = slice(1, None)
    idx2[dim] = slice(None, -1)

    array_out = 0.5 * (array[idx1] + array[idx2])
    return array_out


def cube_interp_extendby1(cube_in, coord):
    import dask.array as da

    dim = cube_in.coord_dims(coord)[0]
    cube_data = cube_in.core_data()
    ndim = cube_in.ndim
    idx1 = [slice(None)] * (ndim)
    idx2 = [slice(None)] * (ndim)
    idx_start = [slice(None)] * (ndim)
    idx_end = [slice(None)] * (ndim)
    idx1[dim] = slice(1, None)
    idx2[dim] = slice(0, -1)
    idx_start[dim] = slice(1)
    idx_end[dim] = slice(-2, -1)
    cube_data[idx_start]
    cube_data[idx1]
    cube_data[idx2]
    cube_data[idx_end]
    array_out = da.concatenate(
        (
            cube_data[idx_start],
            0.5 * (cube_data[idx1] + cube_data[idx2]),
            cube_data[idx_end],
        ),
        axis=dim,
    )

    #    idx1[dim]=slice(1)
    #    array_out=np.concatenate((cube_data[idx1],cube_data),axis=dim)
    return array_out


def cube_interp_reduceby1(cube_in, coord):
    dim = cube_in.coord_dims(coord)
    ndim = cube_in.ndim
    idx1 = [slice(None)] * (ndim)
    idx2 = [slice(None)] * (ndim)
    idx1[dim] = slice(1, None)
    idx2[dim] = slice(None, -1)

    cube_out = 0.5 * (cube_in[idx1] + cube_in[idx2].core_data())
    return cube_out


def load_sum(filename, list_variables, **kwargs):
    cube_out = load(filename, list_variables[0], **kwargs)
    for variable in list_variables[1:]:
        cube_out = cube_out + load(filename, variable, **kwargs)
    return cube_out


def remove_all_coordinates(variable_cube):
    for coordinate in variable_cube.coords():
        variable_cube.remove_coord(coordinate.name())
    return variable_cube


def replacecoordinates(variable_cube, replace_cube):
    variable_cube_out = replace_cube
    variable_cube_out.data = variable_cube.core_data()
    variable_cube_out.rename(variable_cube.name())
    variable_cube_out.units = variable_cube.units
    variable_cube_out.attributes = {}
    return variable_cube_out


# def addcoordinates(filenames, variable,variable_cube,**kwargs):
##    if 'add_coordinates' in kwargs:
##        add_coordinates=kwargs['add_coordinates']
##    else:
##        add_coordinates=None
##
##    if add_coordinates==None:
##        variable_cube=add_dim_coordinates(filenames, variable,variable_cube,**kwargs)
##    else:
##        variable_cube=add_dim_coordinates(filenames, variable,variable_cube,**kwargs)
##        variable_cube=add_aux_coordinates_1dim(filenames, variable,variable_cube,**kwargs)
#    variable_cube=add_dim_coordinates(filenames, variable,variable_cube,**kwargs)
#    variable_cube=add_aux_coordinates_1dim(filenames, variable,variable_cube)
#    variable_cube=add_aux_coordinates_1dim(filenames, variable,variable_cube)
#    return variable_cube
#
# def add_time_coordinate(filenames, variable,variable_cube):
#    time=make_time_coord(filenames)
#    variable_cube.add_dim_coord(time,variable_cube.coord_dims('XTIME')[0])
#    return variable_cube


# def add_dim_coordinates(filenames, variable,variable_cube,add_coordinates=None):
#    from netCDF4 import Dataset  # http://code.google.com/p/netcdf4-python/
#    from iris import load_cube
#    variable_cube_dim= load_cube(filenames, variable)
#
#    attributes=variable_cube_dim.attributes
#    nc_id=Dataset(filenames)
#    nc_variable=nc_id.variables[variable]
#    variable_dimensions=nc_variable.dimensions
#    [str(line) for line in variable_dimensions]
#    DX=attributes['DX']
#    DY=attributes['DY']
#    WEST_EAST_PATCH_END_UNSTAG=attributes['WEST-EAST_PATCH_END_UNSTAG']
#    SOUTH_NORTH_PATCH_END_UNSTAG=attributes['SOUTH-NORTH_PATCH_END_UNSTAG']
#    BOTTOM_TOP_PATCH_END_UNSTAG=attributes['BOTTOM-TOP_PATCH_END_UNSTAG']
#    WEST_EAST_PATCH_END_STAG=attributes['WEST-EAST_PATCH_END_STAG']
#    SOUTH_NORTH_PATCH_END_STAG=attributes['SOUTH-NORTH_PATCH_END_STAG']
#    BOTTOM_TOP_PATCH_END_STAG=attributes['BOTTOM-TOP_PATCH_END_STAG']
#    for dim in range(len(variable_dimensions)):
##        if (variable_dimensions[dim]=='Time'):
##           time=make_time_coord(filenames)
##           variable_cube.add_dim_coord(time,dim)
#        if (variable_dimensions[dim]=='west_east'):
#            west_east=make_westeast_coord(DX,WEST_EAST_PATCH_END_UNSTAG)
#            variable_cube.add_dim_coord(west_east,dim)
#        elif (variable_dimensions[dim]=='south_north'):
#           south_north=make_southnorth_coord(DY, SOUTH_NORTH_PATCH_END_UNSTAG)
#           variable_cube.add_dim_coord(south_north,dim)
#        elif (variable_dimensions[dim]=='bottom_top'):
#           bottom_top=make_bottom_top_coordinate(BOTTOM_TOP_PATCH_END_UNSTAG)
#           variable_cube.add_dim_coord(bottom_top,dim)
#           model_level_number=make_model_level_number_coordinate(BOTTOM_TOP_PATCH_END_UNSTAG)
#           variable_cube.add_aux_coord(model_level_number,dim)
#        elif variable_dimensions[dim]=='west_east_stag':
#           west_east_stag=make_westeast_stag_coord(DX,WEST_EAST_PATCH_END_STAG)
#           variable_cube.add_dim_coord(west_east_stag,dim)
#        elif variable_dimensions[dim]=='south_north_stag':
#           south_north_stag=make_southnorth_stag_coord(DY, SOUTH_NORTH_PATCH_END_STAG)
#           variable_cube.add_dim_coord(south_north_stag,dim)
#        elif variable_dimensions[dim]=='bottom_top_stag':
#           bottom_top_stag=make_bottom_top_stag_coordinate(BOTTOM_TOP_PATCH_END_STAG)
#           variable_cube.add_dim_coord(bottom_top_stag,dim)
#           model_level_number=make_model_level_number_coordinate(BOTTOM_TOP_PATCH_END_STAG)
#           variable_cube.add_aux_coord(model_level_number,dim)
#
#    return variable_cube

# def add_aux_coordinates_1dim(filenames, variable,variable_cube):#,add_coordinates=None):
#    from netCDF4 import Dataset  # http://code.google.com/p/netcdf4-python/
#    from iris import load_cube
#
#    variable_cube_dim= load_cube(filenames, variable)
#    attributes=variable_cube_dim.attributes
#    nc_id=Dataset(filenames)
#
#    nc_variable=nc_id.variables[variable]
#    variable_dimensions=nc_variable.dimensions
#    [str(line) for line in variable_dimensions]
#    DX=attributes['DX']
#    DY=attributes['DY']
#    WEST_EAST_PATCH_END_UNSTAG=attributes['WEST-EAST_PATCH_END_UNSTAG']
#    SOUTH_NORTH_PATCH_END_UNSTAG=attributes['SOUTH-NORTH_PATCH_END_UNSTAG']
#    #BOTTOM_TOP_PATCH_END_UNSTAG=attributes['BOTTOM-TOP_PATCH_END_UNSTAG']
#    WEST_EAST_PATCH_END_STAG=attributes['WEST-EAST_PATCH_END_STAG']
#    SOUTH_NORTH_PATCH_END_STAG=attributes['SOUTH-NORTH_PATCH_END_STAG']
#    #BOTTOM_TOP_PATCH_END_STAG=attributes['BOTTOM-TOP_PATCH_END_STAG']
#    coord_system=make_coord_system(attributes)
#    coords=variable_cube.coords()
##    if type(add_coordinates)!=list:
##        add_coordinates1=add_coordinates
##        add_coordinates=[]
##        add_coordinates.append(add_coordinates1)
##    for coordinate in add_coordinates:
##        if coordinate=='xy':
#    for dim in range(len(coords)):
#        if (coords[dim].name()=='west_east'):
#            projection_x_coord=make_x_coord(DX,WEST_EAST_PATCH_END_UNSTAG,coord_system=coord_system)
#            variable_cube.add_aux_coord(projection_x_coord,dim)
#            x_coord=AuxCoord(variable_cube.coord('west_east').points,long_name='x',units=1)
#            variable_cube.add_aux_coord(x_coord,data_dims=variable_cube.coord_dims('west_east'))
#
#        elif (coords[dim].name()=='south_north'):
#            projection_y_coord=make_y_coord(DY, SOUTH_NORTH_PATCH_END_UNSTAG,coord_system=coord_system)
#            variable_cube.add_aux_coord(projection_y_coord,dim)
#            y_coord=AuxCoord(variable_cube.coord('south_north').points,long_name='y',units=1)
#            variable_cube.add_aux_coord(y_coord,data_dims=variable_cube.coord_dims('south_north'))
#
#        elif (coords[dim].name()=='west_east_stag'):
#            projection_x_stag_coord=make_x_stag_coord(DX,WEST_EAST_PATCH_END_STAG,coord_system=coord_system)
#            variable_cube.add_aux_coord(projection_x_stag_coord,dim)
#            x_coord=AuxCoord(variable_cube.coord('west_east_stag').points,long_name='x',units=1)
#            variable_cube.add_aux_coord(x_coord,data_dims=variable_cube.coord_dims('west_east_stag'))
#
#        elif coords[dim].name()=='south_north_stag':
#            projection_y_stag_coord=make_y_stag_coord(DY, SOUTH_NORTH_PATCH_END_STAG,coord_system=coord_system)
#            variable_cube.add_aux_coord(projection_y_stag_coord,dim)
#            y_coord=AuxCoord(variable_cube.coord('south_north_stag').points,long_name='y',units=1)
#            variable_cube.add_aux_coord(y_coord,data_dims=variable_cube.coord_dims('south_north_stag'))
#
#    return variable_cube
#


def add_aux_coordinates_multidim(
    filenames, variable_cube, add_coordinates=None, constraint=None, **kwargs
):
    from copy import deepcopy

    coords = variable_cube.coords()
    add_coordinates_new = deepcopy(add_coordinates)

    for entry in ["z", "p", "latlon", "zp", "pz"]:
        if entry in add_coordinates_new:
            add_coordinates_new.remove(entry)

    # if constraint in kwargs:
    #     constraint=kwargs.pop('constraint')

    if type(add_coordinates) != list:
        add_coordinates1 = add_coordinates
        add_coordinates = []
        add_coordinates.append(add_coordinates1)

    for coordinate in add_coordinates:
        if coordinate == "z":
            if (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east"
            ):
                z_coord = make_z_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "bottom_top"
                and coords[1].name() == "south_north"
                and coords[2].name() == "west_east"
            ):
                z_coord = make_z_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east_stag"
            ):
                z_coord = make_z_xstag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "bottom_top"
                and coords[1].name() == "south_north"
                and coords[2].name() == "west_east_stag"
            ):
                z_coord = make_z_xstag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north_stag"
                and coords[3].name() == "west_east"
            ):
                z_coord = make_z_ystag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "bottom_top"
                and coords[1].name() == "south_north_stag"
                and coords[2].name() == "west_east"
            ):
                z_coord = make_z_ystag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top_stag"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east"
            ):
                z_stag_coord = make_z_stag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_stag_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "bottom_top_stag"
                and coords[1].name() == "south_north"
                and coords[2].name() == "west_east"
            ):
                z_stag_coord = make_z_stag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_stag_coord, (0, 1, 2))
            else:
                raise ValueError("no z coordinates added")

        if coordinate == "pressure":
            if (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east"
            ):
                p_coord = make_p_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east_stag"
            ):
                p_coord = make_p_xstag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord)
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north_stag"
                and coords[3].name() == "west_east"
            ):
                p_coord = make_p_ystag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top_stag"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east"
            ):
                p_coord = make_p_stag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            else:
                raise ValueError("no p coordinates added")

        if coordinate == "zp" or coordinate == "pz":
            if (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east"
            ):
                z_coord = make_z_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2, 3))
                p_coord = make_p_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east_stag"
            ):
                z_coord = make_z_xstag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2, 3))
                p_coord = make_p_xstag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top"
                and coords[2].name() == "south_north_stag"
                and coords[3].name() == "west_east"
            ):
                z_coord = make_z_ystag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_coord, (0, 1, 2, 3))
                p_coord = make_p_ystag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            elif (
                coords[0].name() == "time"
                and coords[1].name() == "bottom_top_stag"
                and coords[2].name() == "south_north"
                and coords[3].name() == "west_east"
            ):
                z_stag_coord = make_z_stag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(z_stag_coord, (0, 1, 2, 3))
                p_coord = make_p_stag_coordinate(
                    filenames,
                    add_coordinates=add_coordinates_new,
                    constraint=None,
                    **kwargs,
                )
                variable_cube.add_aux_coord(p_coord, (0, 1, 2, 3))
            else:
                raise ValueError("no z and p and lat/lon coordinates added")
    return variable_cube


def make_time_coord(filenames):
    from datetime import datetime

    from cf_units import CALENDAR_STANDARD, date2num
    from iris import coords, load_cube

    Times = load_cube(filenames, "Times")
    filetimes = Times.data
    #    filetimelist = []   # Will contain list of times in seconds since model start time in file.
    timeobjlist = []  # Will contain list of corresponding datetime objects
    for i, filetime in enumerate(filetimes):
        timeobj = datetime.strptime(
            filetime.tostring().decode("UTF-8"), "%Y-%m-%d_%H:%M:%S"
        )
        #        if i == 0:
        #            timeobj0 = timeobj
        #        time_dt = timeobj-timeobj0 # timedelta object representing difference in time from start time
        #        filetimelist.append(time_dt.seconds)
        timeobjlist.append(timeobj)
    #    time_days=empty(len(timeobjlist))
    #    if timeobjlist[0]==datetime(1,1,1,0,0,0):
    #        for i,time in enumerate(timeobjlist):
    #            timeobjlist[i]=timeobjlist[i].replace(year=2000)
    #    #Include a different base_date for dates close to 0001-01-01 (idealised simulations)

    # if timeobjlist[0]<datetime(100,1,1):
    #     base_date=datetime(1,1,1)
    # else:
    base_date = datetime(1970, 1, 1)
    time_units = "days since " + base_date.strftime("%Y-%m-%d")
    #    for i in range(len(timeobjlist)):
    #        time_days[i]=(timeobjlist[i] - base_date).total_seconds() / timedelta(1).total_seconds()

    time_days = date2num(timeobjlist, time_units, CALENDAR_STANDARD)
    time_coord = coords.DimCoord(
        time_days,
        standard_name="time",
        long_name="time",
        var_name="time",
        units=time_units,
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return time_coord


def make_westeast_coord(DX, WEST_EAST_PATCH_END_UNSTAG):
    from iris import coords
    from numpy import arange

    WEST_EAST = arange(0, WEST_EAST_PATCH_END_UNSTAG)
    west_east = coords.DimCoord(
        WEST_EAST,
        standard_name=None,
        long_name="west_east",
        var_name="west_east",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return west_east


def make_westeast_stag_coord(DX, WEST_EAST_PATCH_END_STAG):
    from iris import coords
    from numpy import arange

    WEST_EAST_U = arange(0, WEST_EAST_PATCH_END_STAG)
    west_east_stag = coords.DimCoord(
        WEST_EAST_U,
        standard_name=None,
        long_name="west_east_stag",
        var_name="west_east_stag",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return west_east_stag


def make_southnorth_coord(DY, SOUTH_NORTH_PATCH_END_UNSTAG):
    from iris import coords
    from numpy import arange  # DY=attributes['DY']

    # SOUTH_NORTH_PATCH_END_UNSTAG=attributes['SOUTH-NORTH_PATCH_END_UNSTAG']
    SOUTH_NORTH = arange(0, SOUTH_NORTH_PATCH_END_UNSTAG)
    south_north = coords.DimCoord(
        SOUTH_NORTH,
        standard_name=None,
        long_name="south_north",
        var_name="south_north",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return south_north


def make_southnorth_stag_coord(DY, SOUTH_NORTH_PATCH_END_STAG):
    from iris import coords
    from numpy import arange

    SOUTH_NORTH_V = arange(0, SOUTH_NORTH_PATCH_END_STAG)
    south_north_stag = coords.DimCoord(
        SOUTH_NORTH_V,
        standard_name=None,
        long_name="south_north_stag",
        var_name="south_north_stag",
        units="1",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return south_north_stag


def make_bottom_top_coordinate(BOTTOM_TOP_PATCH_END_UNSTAG):
    from iris import coords
    from numpy import arange

    BOTTOM_TOP = arange(0, BOTTOM_TOP_PATCH_END_UNSTAG)
    bottom_top = coords.DimCoord(
        BOTTOM_TOP,
        standard_name=None,
        long_name="altitude",
        var_name="altitude",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return bottom_top


def make_bottom_top_stag_coordinate(BOTTOM_TOP_PATCH_END_STAG):
    from iris import coords
    from numpy import arange

    BOTTOM_TOP_W = arange(0, BOTTOM_TOP_PATCH_END_STAG)
    bottom_top_stag = coords.DimCoord(
        BOTTOM_TOP_W,
        standard_name=None,
        long_name="altitude_stag",
        var_name="altitude_stag",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
        circular=False,
    )
    return bottom_top_stag


def make_model_level_number_coordinate(BOTTOM_TOP_PATCH_END):
    from iris import coords
    from numpy import arange

    MODEL_LEVEL_NUMBER = arange(0, BOTTOM_TOP_PATCH_END)
    model_level_number = coords.AuxCoord(
        MODEL_LEVEL_NUMBER, standard_name="model_level_number", units="1"
    )
    return model_level_number


def make_coord_system(attributes):
    from iris import coord_systems

    #    :CEN_LAT = -3.212929f ;
    # 		:CEN_LON = -60.59799f ;
    # 		:TRUELAT1 = 0.f ;
    # 		:TRUELAT2 = -5.f ;
    # 		:MOAD_CEN_LAT = -3.212929f ;
    # 		:STAND_LON = -60.f ;
    # 		:POLE_LAT = 90.f ;
    # 		:POLE_LON = 0.f ;
    # 		:GMT = 0.f ;
    # 		:JULYR = 2014 ;
    # 		:JULDAY = 244 ;
    # 		:MAP_PROJ = 1 ;
    # 		:MAP_PROJ_CHAR = "Lambert Conformal" ;
    MAP_PROJ_CHAR = attributes["MAP_PROJ_CHAR"]
    # MAP_PROJ = attributes["MAP_PROJ"] # not sure what the MAP_PROJ is for, it looks like an integer which also specifies the coordinate system

    # cartesian coordinate system (idealized simulations):
    if MAP_PROJ_CHAR == "Cartesian":  # and MAP_PROJ == 0:
        coord_system = None

    # lambert Conformal system (idealized simulations):
    if MAP_PROJ_CHAR == "Lambert Conformal":  # and MAP_PROJ == 1:
        CEN_LON = attributes["CEN_LON"]
        TRUELAT1 = attributes["TRUELAT1"]
        TRUELAT2 = attributes["TRUELAT2"]
        MOAD_CEN_LAT = attributes["MOAD_CEN_LAT"]
        STAND_LON = attributes["STAND_LON"]
        POLE_LAT = attributes["POLE_LAT"]
        POLE_LON = attributes["POLE_LON"]
        coord_system = coord_systems.LambertConformal(
            central_lat=MOAD_CEN_LAT,
            central_lon=CEN_LON,
            false_easting=0.0,
            false_northing=0.0,
            secant_latitudes=(TRUELAT1, TRUELAT2),
        )
    if MAP_PROJ_CHAR == "Polar-Stereographic":
        CEN_LAT = attributes["CEN_LAT"]
        CEN_LON = attributes["CEN_LON"]
        coord_system = coord_systems.PolarStereographic(
            central_lat=CEN_LAT,
            central_lon=CEN_LON,
            false_easting=0.0,  # I am not sure of any of these, but for now leave them as standard
            false_northing=0.0,
            true_scale_lat=None,
            scale_factor_at_projection_origin=None,
            ellipsoid=None,
        )
    return coord_system


def make_x_coord(DX, WEST_EAST_PATCH_END_UNSTAG, coord_system):
    from iris import coords
    from numpy import arange, array, transpose

    X = DX * (arange(0, WEST_EAST_PATCH_END_UNSTAG) + 0.5)
    bounds = transpose(
        array(
            [
                DX * (arange(0, WEST_EAST_PATCH_END_UNSTAG)),
                DX * (arange(0, WEST_EAST_PATCH_END_UNSTAG) + 1),
            ]
        )
    )
    x_coord = coords.AuxCoord(
        X,
        standard_name="projection_x_coordinate",
        long_name="x",
        var_name="x",
        units="m",
        bounds=bounds,
        attributes=None,
        coord_system=coord_system,
    )
    # x_coord.add_dim_coord(west_east,0)
    return x_coord


def make_x_stag_coord(DX, WEST_EAST_PATCH_END_STAG, coord_system=None):
    from iris import coords
    from numpy import arange

    X_U = DX * (arange(0, WEST_EAST_PATCH_END_STAG) - 1)
    x_stag_coord = coords.AuxCoord(
        X_U,
        standard_name="projection_x_coordinate",
        long_name="x",
        var_name="x",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=coord_system,
    )
    # x_stag_coord.add_dim_coord(west_east_stag,0)
    return x_stag_coord


def make_y_coord(DY, SOUTH_NORTH_PATCH_END_UNSTAG, coord_system=None):
    from iris import coords
    from numpy import arange, array, transpose  # DY=attributes['DY']

    Y = DY * (arange(0, SOUTH_NORTH_PATCH_END_UNSTAG) + 0.5)
    bounds = transpose(
        array(
            [
                DY * (arange(0, SOUTH_NORTH_PATCH_END_UNSTAG)),
                DY * (arange(0, SOUTH_NORTH_PATCH_END_UNSTAG) + 1),
            ]
        )
    )
    y_coord = coords.AuxCoord(
        Y,
        standard_name="projection_y_coordinate",
        long_name="y",
        var_name="y",
        units="m",
        bounds=bounds,
        attributes=None,
        coord_system=coord_system,
    )
    # y_coord.add_dim_coord(south_north,0)
    return y_coord


def make_y_stag_coord(DY, SOUTH_NORTH_PATCH_END_STAG, coord_system=None):
    from iris import coords
    from numpy import arange

    Y_V = DY * (arange(0, SOUTH_NORTH_PATCH_END_STAG) - 1)
    y_stag_coord = coords.AuxCoord(
        Y_V,
        standard_name="projection_y_coordinate",
        long_name="y",
        var_name="y",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=coord_system,
    )
    # y_stag_coord.add_dim_coord(south_north_stag,0)
    return y_stag_coord


def make_z_coordinate(filenames, **kwargs):
    from iris import coords

    z = calculate_rams_geopotential_height(filenames, **kwargs)
    z_coord = coords.AuxCoord(
        z.core_data(),
        standard_name="geopotential_height",
        long_name="geopotential_height",
        var_name="z",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return z_coord


def make_z_xstag_coordinate(filenames, **kwargs):
    from iris import coords

    z = calculate_rams_geopotential_height_xstag(filenames, **kwargs)
    z_coord = coords.AuxCoord(
        z,
        standard_name="geopotential_height",
        long_name="geopotential_height",
        var_name="z",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return z_coord


def make_z_ystag_coordinate(filenames, **kwargs):
    from iris import coords

    z = calculate_rams_geopotential_height_ystag(filenames, **kwargs)
    z_coord = coords.AuxCoord(
        z,
        standard_name="geopotential_height",
        long_name="geopotential_height",
        var_name="z",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return z_coord


def make_z_stag_coordinate(filenames, **kwargs):
    from iris import coords

    z = calculate_rams_geopotential_height_stag(filenames, **kwargs)
    z_coord = coords.AuxCoord(
        z.core_data(),
        standard_name="geopotential_height",
        long_name="z",
        var_name="z",
        units="m",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return z_coord


def make_p_coordinate(filenames, **kwargs):
    from iris import coords

    p = calculate_rams_pressure(filenames, **kwargs)
    p_coord = coords.AuxCoord(
        p.core_data(),
        standard_name=None,
        long_name="pressure",
        var_name="pressure",
        units="Pa",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return p_coord


def make_p_xstag_coordinate(filenames, **kwargs):
    from iris import coords

    p = calculate_rams_pressure_xstag(filenames, **kwargs)
    p_coord = coords.AuxCoord(
        p.core_data(),
        standard_name=None,
        long_name="pressure",
        var_name="pressure",
        units="Pa",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return p_coord


def make_p_ystag_coordinate(filenames, **kwargs):
    from iris import coords

    p = calculate_rams_pressure_ystag(filenames, **kwargs)
    p_coord = coords.AuxCoord(
        p.core_data(),
        standard_name=None,
        long_name="pressure",
        var_name="pressure",
        units="Pa",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return p_coord


def make_p_stag_coordinate(filenames, **kwargs):
    from iris import coords

    p = calculate_rams_pressure_stag(filenames, **kwargs)
    p_coord = coords.AuxCoord(
        p.core_data(),
        standard_name=None,
        long_name="pressure",
        var_name="pressure",
        units="Pa",
        bounds=None,
        attributes=None,
        coord_system=None,
    )
    return p_coord


def collapse_removecoord(cube, coord, aggregator):
    for coordinate in cube.coords():
        if coordinate.ndim > 1 and cube.coord_dims(coord)[0] in cube.coord_dims(
            coordinate
        ):
            cube.remove_coord(coordinate.name())
    cube = cube.collapsed((coord), aggregator)
    return cube



import datetime

import xarray as xr

from .mesonh_calculate_products import mesonh_calculate_brightness_temp
from .TAMS import run


def mesonh_run_tams(mesonh_xarray, CONFIG):
    """
    Inputs:
        mesonh_xarray: xarray Dataset containing MesoNH data calculated from mesonh_load.py
        CONFIG: User configuration file
    Outputs:
        ce: a geopandas dataframe with the identified cloud elements
        latlon_coord_system: a tuple of the latitude and longitude coordinate arrays
    """

    # make a copy of MesoNH dataset to configure it to an acceptable format for TAMS
    mesonh_for_tams_copy = xr.Dataset({})

    # if brightness temperature is already in mesonh_xarray use it
    if "TB" not in mesonh_xarray:
        tb = xr.DataArray(
            mesonh_calculate_brightness_temp(mesonh_xarray), dims=["time", "y", "x"]
        )

    else:
        tb = mesonh_xarray["TB"]

    mesonh_for_tams_copy["ctt"] = tb.assign_attrs(
        {"long_name": "Brightness temperature", "units": "K"}
    )
    mesonh_for_tams_copy["ctt"].chunk(mesonh_xarray["top"].chunksizes)
    mesonh_xarray = mesonh_xarray.assign(TB=mesonh_for_tams_copy["ctt"])

    # # if precipitation rate is already in mesonh_xarray use it
    # if "PR" not in mesonh_xarray:
    #     pr = mesonh_calculate_precip_rate(mesonh_xarray)
    #     mesonh_for_tams_copy["pr"] = pr.assign_attrs(
    #         {"long_name":  "Precipitation rate",
    #         "units":      "mm h-1"}
    #     )
    #     mesonh_for_tams_copy["pr"].chunk(mesonh_xarray["TOPT"].chunksizes)
    # else:
    #     pr = mesonh_xarray["PR"]

    # format the times into a list of datetime objects
    datetime_times = [
        datetime.datetime.strptime(str(time), "%Y-%m-%dT%H:%M:%S.%f000")
        for time in mesonh_xarray["time"].values
    ]

    # make the coordinates and rename the dimensions
    mesonh_for_tams_copy = mesonh_for_tams_copy.assign_coords(
        time=(["time"], datetime_times)
    )
    mesonh_for_tams_copy = mesonh_for_tams_copy.assign_coords(
        lat=(["time", "y", "x"], mesonh_xarray["lat"].data)
    )
    mesonh_for_tams_copy = mesonh_for_tams_copy.assign_coords(
        lon=(["time", "y", "x"], mesonh_xarray["lon"].data)
    )

    mesonh_for_tams_copy = mesonh_for_tams_copy.rename(
        {"y": "lat", "x": "lon"}
    )  # rename the dimensions

    ce, latlon_coord_system = run(mesonh_for_tams_copy, **CONFIG["mesonh"]["tams"])

    return ce, latlon_coord_system



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 28 15:00:59 2024

@author: thahn
"""

# =============================================================================
# This contains all of the functions to calculate basic bulk cell statistics (ETH, area, etc.) from CoCoMET-US data
# =============================================================================

import logging
from copy import deepcopy

import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm


# Calculate nearest item in list to given pivot
def find_nearest(array: np.ndarray, pivot) -> int:
    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def calculate_var_max_height(
    analysis_object: dict,
    threshold: float,
    variable: str | None = None,
    cell_footprint_height: float = 2,
    quantile: float = 0.95,
    **args: dict,
) -> pd.DataFrame | None:
    """


    Parameters
    ----------
    analysis_object : dict
        A  CoMET-US standard analysis object containing at least US_tracks and US_segmentation_2d or US_segmentation_3d, and segmentation_xarray.
    threshold : float
        The value which needs to be exceeded to count towards the var top height. I.e. 15 for reflectivity.
    variable : str, optional
        The variable from the input segmentation_xarray which should be used for calculating var_max_height. The default is None.
    cell_footprint_height : float, optional
        The height used to calculate the cell area to determine where to calculate var_max_heights. The default is 2km.
    quantile : float, optional
        The percentile of calculated max heights to return. The default is 0.95.
    **args : dict
        Throw away inputs.

    Raises
    ------
    Exception
        Exception if missing segmentation data from the analysis object.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, max_height in km.

    """

    # If input variable field is 2D return None. Also, if DataArray, use those values for calculations. If Dataset, use tracking_var to get variable
    if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
        if len(analysis_object["segmentation_xarray"].shape) != 4:
            logging.warning("!=====Input Variable is not 3D=====!")
            return None

        variable_field = analysis_object["segmentation_xarray"]

    else:
        if len(analysis_object["segmentation_xarray"][variable].shape) != 4:
            logging.warning("!=====Input Variable is not 3D=====!")
            return None

        variable_field = analysis_object["segmentation_xarray"][variable]

    # If 3D segmentation is available, use that to calculate cell footprint, otherwise use 2D segmentation
    if analysis_object["US_segmentation_3d"] is not None:
        height_index = find_nearest(
            analysis_object["US_segmentation_3d"].altitude.values,
            cell_footprint_height * 1000,
        )

        footprint_data = analysis_object["US_segmentation_3d"].Feature_Segmentation[
            :, height_index
        ]

    elif analysis_object["US_segmentation_2d"] is not None:
        footprint_data = analysis_object["US_segmentation_2d"].Feature_Segmentation

    else:
        raise ValueError("!=====Missing Segmentation Input=====!")

    # Pull out your tracks DataFrame
    Tracks = analysis_object["US_tracks"]

    # Build lookup dicts
    features_by_frame = (Tracks.groupby("frame")["feature_id"].apply(list).to_dict())
    cell_by_frame_feat = (Tracks.groupby(["frame", "feature_id"])["cell_id"].min().to_dict())

    # Convert DataArrays to NumPy once
    footprint_np = footprint_data.values      
    var_np = variable_field.values      
    altitudes = variable_field.altitude.values  

    # Prepare the output dict
    max_height_info = {
        "frame": [],
        "feature_id": [],
        "cell_id": [],
        "max_height": [],
    }  # in km

    frame_groups = analysis_object["US_tracks"].groupby("frame")

    # Loop over frames → features
    for frame in tqdm(
        features_by_frame.keys(),
        desc=f"=====Calculating Max Heights=====",
        total=len(features_by_frame),
    ):
        for feature_id in features_by_frame[frame]:
            # get the precomputed cell_id
            cell_id = cell_by_frame_feat[(frame, feature_id)]

            # find all (y,x) pixels in this footprint
            idxs = np.argwhere(footprint_np[frame] == feature_id)
            if idxs.size == 0:
                max_height = np.nan
            else:
                # for each footprint pixel, find the highest altitude index > threshold
                heights = []
                for iy, ix in idxs:
                    col = var_np[frame, :, iy, ix] 
                    above = np.where(col > threshold)[0]
                    if above.size:
                        max_idx = above.max()
                        heights.append(altitudes[max_idx])
                    else:
                        heights.append(np.nan)

                # reduce by quantile (and convert to km)
                if np.all(np.isnan(heights)):
                    max_height = np.nan
                else:
                    max_height = np.nanquantile(heights, quantile) / 1000

            # append results
            max_height_info["frame"].append(frame)
            max_height_info["feature_id"].append(feature_id)
            max_height_info["cell_id"].append(cell_id)
            max_height_info["max_height"].append(max_height)

    # Return a DataFrame
    return pd.DataFrame(max_height_info)


def calculate_max_intensity(
    analysis_object: dict,
    variable: str | None = None,
    cell_footprint_height: float | None = None,
    **args,
) -> pd.DataFrame | None:
    """


    Parameters
    ----------
    analysis_object : dict
        A  CoCoMET-US standard analysis object containing at least US_tracks and US_segmentation_2d or US_segmentation_3d, and segmentation_xarray.
    variable : str, optional
        The variable from the input segmentation_xarray which should be used for calculating var_max_height. The default is None.
    cell_footprint_height : float, optional
        The height used to determine the location of which cells to calculate the max_intensity of. If None, will calculate the max_intensity of all cells. The default is None.
    quantile : float, optional
        The percentile of calculated max heights to return. The default is 0.95.
    **args : dict
        Throw away inputs.

    Raises
    ------
    Exception
        Exception if missing segmentation data from the analysis object.
        Exception if there is more than one frame associated with a feature id

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, max_intensity where max_intensity is in the unit of the variable.

    """

    # If DataArray, use those values for calculations. If Dataset, use tracking_var to get variable
    if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
        variable_field = analysis_object["segmentation_xarray"]

    else:
        variable_field = analysis_object["segmentation_xarray"][variable]

    # If 3D segmentation is available, use that to calculate cell footprint, otherwise use 2D segmentation
    if analysis_object["US_segmentation_3d"] is not None:
        segmentation = analysis_object["US_segmentation_3d"].Feature_Segmentation
        segmentation = analysis_object["US_segmentation_3d"].Feature_Segmentation

        if cell_footprint_height is not None:
            height_index = find_nearest(
                analysis_object["US_segmentation_3d"].altitude.values,
                cell_footprint_height * 1000,
            )

            features_across_footprint = np.unique(segmentation[:, height_index])[1:]

        else:
            features_across_footprint = np.unique(segmentation)[1:]

    elif analysis_object["US_segmentation_2d"] is not None:
        segmentation = analysis_object["US_segmentation_2d"].Feature_Segmentation
        features_across_footprint = np.unique(segmentation)[1:]

    else:
        raise ValueError("!=====Missing Segmentation Input=====!")

    max_intensity_info = {
        "frame": [],
        "feature_id": [],
        "cell_id": [],
        "max_intensity": [],
    }
    Tracks = analysis_object["US_tracks"]

    # Lookup dict to avoid groupby
    frames_by_feat = dict(zip(Tracks["feature_id"], Tracks["frame"]))
    cells_by_feat  = dict(zip(Tracks["feature_id"], Tracks["cell_id"]))

    # Convert DataArrays to plain NumPy arrays once
    seg_np = segmentation.values         # shape (n_frames, H, W)
    var_np = variable_field.values       # same shape

    # Loop over each frame, then apply a mask for the cell/feature, then within that find the highest val
    for feature_id, frame in tqdm(
        frames_by_feat.items(),
        desc="=====Calculating Max Intensity=====",
        total=len(frames_by_feat),
    ):
        # If feature covers your footprint set, compute its max; else NaN
        if feature_id in features_across_footprint:
            # mask & reduce in pure NumPy
            one_feat = np.where(
                seg_np[frame] == feature_id,
                var_np[frame],
                0,
            )
            max_val = np.nanmax(one_feat)
        else:
            max_val = np.nan

        # Append straight from our dicts no more groupby inside
        max_intensity_info["frame"].append(frame)
        max_intensity_info["feature_id"].append(feature_id)
        max_intensity_info["cell_id"].append(cells_by_feat[feature_id])
        max_intensity_info["max_intensity"].append(max_val)

    return pd.DataFrame(max_intensity_info)


def calculate_area(
    analysis_object: dict,
    height: float = 2,
    variable: str | None = None,
    threshold: float | None = None,
    **args: dict,
) -> pd.DataFrame:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks and US_segmentation_2d or US_segmentation_3d.
    height : float, optional
        The height which is used to calculate the area of cells. The default is 2km.
    variable : str, optional
        Variable to which we should apply the threshold. The default is None.
    threshold : float, optional
        Value of which the area should be greater than. The default is None.
    **args : dict
        Throw away variables.

    Raises
    ------
    Exception
        Exception if missing segmentation input from the analysis object.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, area where area is in km^2.

    """

    # If 3D segmentation is available, use that at given height, otherwise use 2D segmentation
    if analysis_object["US_segmentation_3d"] is not None:
        height_index = find_nearest(
            analysis_object["US_segmentation_3d"].altitude.values, height * 1000
        )

        mask = analysis_object["US_segmentation_3d"].Feature_Segmentation[
            :, height_index
        ]

    elif analysis_object["US_segmentation_2d"] is not None:
        mask = analysis_object["US_segmentation_2d"].Feature_Segmentation

    else:
        raise ValueError("!=====Missing Segmentation Input=====!")

    # Enforce threshold if present
    if threshold is not None:
        # Load background varaible
        if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
            variable_field = analysis_object["segmentation_xarray"]
        else:
            variable_field = analysis_object["segmentation_xarray"][variable]

        if len(variable_field.shape) == 4:
            height_index = find_nearest(
                analysis_object["segmentation_xarray"]["altitude"].values, height * 1000
            )

            variable_field = variable_field[:, height_index]

        mask.values[variable_field.values < threshold] = -1

    area_info = {"frame": [], "feature_id": [], "cell_id": [], "area": []}  # in km^2

    # We first calculate the area of each individual cell
    # First get the size of each dimension
    x_dim_sizes = np.abs(np.diff(mask.projection_x_coordinate))
    y_dim_sizes = np.abs(np.diff(mask.projection_y_coordinate))

    # These are one cell too small due to how diff works, so infer last cell size using the same cell size as the previous cell
    # x_dim_sizes.append(x_dim_sizes[-1])
    x_dim_sizes = np.append(x_dim_sizes, x_dim_sizes[-1])
    y_dim_sizes = np.append(y_dim_sizes, y_dim_sizes[-1])

    # Multiply each cell by the other to get an area for one individual cell
    cell_areas = np.outer(y_dim_sizes, x_dim_sizes)

    frame_groups = analysis_object["US_tracks"].groupby("frame")

    # Loop over frames
    for ii, frame in tqdm(
        enumerate(frame_groups),
        desc="=====Calculating Areas=====",
        total=frame_groups.ngroups,
    ):
        # Loop over each feature
        for feature in frame[1].groupby("feature_id"):
            # Get valid indices of a given features
            proper_indices = np.argwhere(mask[frame[0]].values == feature[0])

            # Cells which have no segmented output should get a NaN
            if len(proper_indices) == 0:
                area_info["frame"].append(frame[0])
                area_info["feature_id"].append(feature[0])
                area_info["cell_id"].append(feature[1]["cell_id"].min())
                area_info["area"].append(np.nan)
                continue

            # Sum up all the areas that comprise it
            total = np.sum([cell_areas[iy, ix] for iy, ix in proper_indices])

            # Push info to dictionary
            area_info["frame"].append(frame[0])
            area_info["feature_id"].append(feature[0])
            area_info["cell_id"].append(feature[1]["cell_id"].min())
            area_info["area"].append(total / 1e6)

    return pd.DataFrame(area_info)


def calculate_volume(
    analysis_object: dict,
    variable: str | None = None,
    threshold: float | None = None,
    **args: dict,
) -> pd.DataFrame:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks and US_segmentation_3d.
    variable : str, optional
        Variable to which we should apply the threshold. The default is None.
    threshold : float, optional
        Value of which the area should be greater than. The default is None.
    **args : dict
        Throw away variables.

    Raises
    ------
    Exception
        Exception if missing segmentation input from the analysis object.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, volume where area is in km^3.

    """

    if analysis_object["US_segmentation_3d"] is None:
        raise ValueError(
            "!=====3D Segmentation Data is Required for Volume Calculation=====!"
        )

    mask = analysis_object["US_segmentation_3d"].Feature_Segmentation

    if threshold is not None:
        # Load background varaible
        if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
            variable_field = analysis_object["segmentation_xarray"]
        else:
            variable_field = analysis_object["segmentation_xarray"][variable]

        if len(variable_field.shape) != 4:
            raise ValueError(
                "=====Must have a 3D segmentation_xarray for threshold application====="
            )

        # Enforce threshold
        mask.values[variable_field.values < threshold] = -1

    volume_info = {
        "frame": [],
        "feature_id": [],
        "cell_id": [],
        "volume": [],  # in km^3
    }

    # We first calculate the area of each individual cell
    # First get the size of each dimension
    x_dim_sizes = np.diff(mask.projection_x_coordinate)
    y_dim_sizes = np.diff(mask.projection_y_coordinate)
    z_dim_sizes = np.diff(mask.altitude)

    # These are one cell too small due to how diff works, so infer last cell size using the same cell size as the previous cell
    # x_dim_sizes.append(x_dim_sizes[-1])
    x_dim_sizes = np.append(x_dim_sizes, x_dim_sizes[-1])
    y_dim_sizes = np.append(y_dim_sizes, y_dim_sizes[-1])
    z_dim_sizes = np.append(z_dim_sizes, z_dim_sizes[-1])

    # use Einstein sum notation to get volume of cells
    cell_volumes = np.einsum("i,j,k->ijk", z_dim_sizes, y_dim_sizes, x_dim_sizes)

    frame_groups = analysis_object["US_tracks"].groupby("frame")

    # Loop over frames
    for ii, frame in tqdm(
        enumerate(frame_groups),
        desc="=====Calculating Volumes=====",
        total=frame_groups.ngroups,
    ):
        # Loop over each feature
        for feature in frame[1].groupby("feature_id"):
            # Get valid indices of a given features
            proper_indices = np.argwhere(mask[frame[0]].values == feature[0])

            # Cells which have no segmented output should get a NaN
            if len(proper_indices) == 0:
                volume_info["frame"].append(frame[0])
                volume_info["feature_id"].append(feature[0])
                volume_info["cell_id"].append(feature[1]["cell_id"].min())
                volume_info["volume"].append(np.nan)
                continue

            # Sum up all the volumes that comprise it
            total = np.sum([cell_volumes[iz, iy, ix] for iz, iy, ix in proper_indices])

            # Push info to dictionary
            volume_info["frame"].append(frame[0])
            volume_info["feature_id"].append(feature[0])
            volume_info["cell_id"].append(feature[1]["cell_id"].min())
            volume_info["volume"].append(total / 1e9)

    return pd.DataFrame(volume_info)


def calculate_velocity(
    analysis_object: dict,
    **args: dict,
) -> pd.DataFrame:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks and tracking_xarray.
    **args : dict
        Throw away variables.

    Raises
    ------
    Exception
        Exception if missing tracks input from the analysis object.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, velocity where velocity is in m/s or m/frame.

    """

    if analysis_object["US_tracks"] is None:
        raise ValueError("!=====Tracks Data is Required for Velocity Calculation=====!")

    # Check if altitude values are present
    dim = 2
    if "up_down" in analysis_object["US_tracks"]:
        if np.isfinite(np.nanmax(analysis_object["US_tracks"]["up_down"])):
            dim = 3

    # Get dt
    if ("tracking_xarray" in analysis_object) and (
        "DT" in analysis_object["tracking_xarray"].attrs
    ):
        dt_from_tracking = analysis_object["tracking_xarray"].attrs["DT"]  # in s
    else:
        dt_from_tracking = 1.0
        logging.warning("Could not find DT in dataset, computing velocity in m/frame")

    def calculate_row_velocity(row1, row2, dt_arr, dim):
        # given 2 rows of the tracks dataframe, compute the distance between the 2 in 3 components:
        x1 = row1["projection_x"]
        x2 = row2["projection_x"]
        y1 = row1["projection_y"]
        y2 = row2["projection_y"]

        # f2 = row1["frame"]
        # f1 = row2["frame"]

        t1 = row1["time"]
        t2 = row2["time"]

        dt = (
            t2 - t1
        ).total_seconds()  # If a cell is not tracked for a frame, account for this (e.x. cell is in frame 1 and 3)

        distancecomps = ((y2 - y1), (x2 - x1))
        distance = np.sqrt(distancecomps[0] ** 2 + distancecomps[1] ** 2)

        velocity = ((y2 - y1) / dt, (x2 - x1) / dt)
        speed = np.sqrt((velocity[0]) ** 2 + (velocity[1]) ** 2)

        if dim == 3:
            z1 = row1["altitude"] * 1000
            z2 = row2["altitude"] * 1000

            distancecomps = ((z2 - z1), (y2 - y1), (x2 - x1))
            distance = np.sqrt(
                distancecomps[0] ** 2 + distancecomps[1] ** 2 + distancecomps[2] ** 2
            )

            velocity = ((z2 - z1) / dt, (y2 - y1) / dt, (x2 - x1) / dt)
            speed = np.sqrt(velocity[0] ** 2 + velocity[1] ** 2 + velocity[2] ** 2)

        return (distancecomps, distance, velocity, speed)

    Tracks = deepcopy(analysis_object["US_tracks"])
    velocity_info = {
        "frame": [],
        "feature_id": [],
        "cell_id": [],
        "velocity": [],  # in m / s
        "speed": [],  # in m / s
    }

    # Make dt into an array
    if type(dt_from_tracking) is list or type(dt_from_tracking) is np.ndarray:
        # Check that the number of frames and the number of dt"s are the same
        dt_array = list(dt_from_tracking)  # append the last value from the list
        dt_array.append(dt_from_tracking[-1])
    else:
        dt_array = np.ones(len(np.unique(Tracks["frame"])) + 1) * dt_from_tracking

    for cell in tqdm(
        Tracks.groupby("cell_id"),
        desc="=====Calculating Cell Velocities=====",
        total=len(np.unique(Tracks.cell_id.values)),
    ):
        for ii in range(len(cell[1].feature_id.values)):
            # check if last row
            if ii == (len(cell[1].feature_id.values) - 1):
                velocity_info["frame"].append(cell[1].iloc[ii].frame)
                velocity_info["feature_id"].append(cell[1].iloc[ii].feature_id)
                velocity_info["cell_id"].append(cell[0])
                velocity_info["velocity"].append(np.nan)
                velocity_info["speed"].append(np.nan)
                continue

            # else calculate speed / velocity based off of next frame
            dat_out = calculate_row_velocity(
                cell[1].iloc[ii], cell[1].iloc[ii + 1], dt_array, dim
            )

            velocity_info["frame"].append(cell[1].iloc[ii].frame)
            velocity_info["feature_id"].append(cell[1].iloc[ii].feature_id)
            velocity_info["cell_id"].append(cell[0])
            velocity_info["velocity"].append(dat_out[-2])
            velocity_info["speed"].append(dat_out[-1])

    # for i in tqdm(
    #     Tracks.index,
    #     desc="=====Calculating Cell Velocities=====",
    #     total=len(Tracks.index),
    # ):
    #     cell_id = Tracks.iloc[i]["cell_id"]
    #     frame = Tracks.iloc[i]["frame"]
    #     feature_id = Tracks.iloc[i]["feature_id"]

    #     velocity_info["frame"].append(frame)
    #     velocity_info["feature_id"].append(feature_id)
    #     velocity_info["cell_id"].append(cell_id)

    #     if Tracks.iloc[i]["frame"] == np.min(Tracks["frame"]) and cell_id != -1:
    #         velocity_info["velocity"].append(tuple(np.zeros(dim)))
    #         velocity_info["speed"].append(0)

    #     # why?????????
    #     elif cell_id not in Tracks[:i]["cell_id"].values:  # check if this is a new cell
    #         # if new, look for closest cell in same frame to copy velocity from
    #         same_frame_index_list = Tracks.index[
    #             Tracks["frame"] == Tracks.iloc[i]["frame"]
    #         ].tolist()
    #         same_frame_index_list.remove(i)
    #         if len(same_frame_index_list) == 0:
    #             velocity_info["velocity"].append(tuple(np.zeros(dim)))
    #             velocity_info["speed"].append(0)
    #             continue

    #         min_distance = np.inf
    #         closest_vel = tuple(np.zeros(dim))
    #         closest_speed = 0
    #         for j in same_frame_index_list:
    #             target_row = Tracks.iloc[j]
    #             if len(velocity_info["velocity"]) < j:
    #                 continue

    #             target_velocity = velocity_info["velocity"][j]
    #             target_speed = velocity_info["speed"][j]

    #             distance = calculate_row_velocity(
    #                 Tracks.iloc[i], target_row, dt_array, dim
    #             )[1]

    #             if distance < min_distance:
    #                 min_distance = distance
    #                 closest_vel = target_velocity
    #                 closest_speed = target_speed

    #         # If there are no other cells in the frame, append 0
    #         velocity_info["velocity"].append(closest_vel)
    #         velocity_info["speed"].append(closest_speed)

    #     else:
    #         # if the cell is not new, determine velocity from the change in position from previous frame
    #         index = np.argwhere(Tracks[:i]["cell_id"].values == cell_id)[-1][
    #             0
    #         ]  # index of previous row with same cell id
    #         velocity_speed = calculate_row_velocity(
    #             Tracks.iloc[i], Tracks.iloc[index], dt_array, dim
    #         )[2:]
    #         velocity_info["velocity"].append(velocity_speed[0])
    #         velocity_info["speed"].append(velocity_speed[1])

    # for i in Tracks.index:
    #     cell_id = Tracks.iloc[i]["cell_id"]
    #     if velocity_info["velocity"][i] == tuple(
    #         np.zeros(dim)
    #     ):  # go back and look at where velocity was set to zero and see if we can extrapolate backwards
    #         index_list = list(np.argwhere(Tracks["cell_id"].values == cell_id))
    #         index_list.remove(i)
    #         if len(index_list) == 0:
    #             continue
    #         velocity_info["velocity"][i] = velocity_info["velocity"][index_list[0][0]]
    #         velocity_info["speed"][i] = velocity_info["speed"][index_list[0][0]]

    velocity_info_df = pd.DataFrame(velocity_info)

    return velocity_info_df


def calculate_cell_growth(
    analysis_object: dict,
    variable: str,
    threshold: float | None = None,
    **args: dict,
) -> pd.DataFrame:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks.
    variable : str
        Variable to which we should apply the threshold.
    threshold : float, optional
        Value of which the area should be greater than. The default is None.
    **args : dict
        Throw away variables.

    Raises
    ------
    Exception
        Exception if missing tracks input from the analysis object.
        Exception if the segmentation data is not 2D or 3D.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, cell_growth where cell_growth is in m^3 / s.

    """

    if analysis_object["US_tracks"] is None:
        raise ValueError("!=====Tracks Data is Required for Velocity Calculation=====!")

    def _variable_lookup(analysis_object, var, **args):

        if var == "var_max_height":
            variable = "max_height"
            return calculate_var_max_height(analysis_object, **args)["max_height"]
        elif var == "max_intensity":
            return calculate_max_intensity(analysis_object, **args)["max_intensity"]
        elif var == "area":
            return calculate_area(analysis_object, **args)["area"]
        elif var == "volume":
            return calculate_volume(analysis_object, **args)["volume"]
        elif var == "velocity" or var == "speed":
            variable = "speed"
            return calculate_velocity(analysis_object, **args)["speed"]
        elif var == "perimeter":
            return calculate_perimeter(analysis_object, **args)["perimeter"]
        elif var == "convexity":
            return calculate_irregularity(analysis_object, **args)["convexity"]
        elif var == "sphericity":
            return calculate_irregularity(analysis_object, **args)["sphericity"]
        else:
            raise KeyError("!=====Invalid variable for cell growth rate=====!")

    # Check if the code is 3D
    if analysis_object["US_segmentation_3d"] is not None:

        dim = 3

        # If there is a given variable use that
        # TODO: there's definitely a better way to do this. in _variable_lookup maybe if variable is None default to either volume or area
        var = _variable_lookup(analysis_object, var=variable, **args)

    # If there is no 3D segmentation check if there is 2D
    elif analysis_object["US_segmentation_2d"] is not None:

        dim = 2

        # If there is a given variable use that
        if variable != None:
            if variable in args:
                var = args[variable][variable]
            else:
                var = _variable_lookup(analysis_object, var=variable, **args)

        # If there is no variable, default to area
        else:
            if "area" in args:
                var = args["area"]["area"]
            else:
                var = _variable_lookup(analysis_object, var="area", **args)

    else:
        raise Exception("!=====Missing Segmentation Input=====!")

    # Get Tracks
    Tracks = analysis_object["US_tracks"]
    Tracks_with_var = deepcopy(Tracks).join(var)

    # Get dt
    if ("tracking_xarray" in analysis_object) and (
        "DT" in analysis_object["tracking_xarray"].attrs
    ):
        dt_from_tracking = analysis_object["tracking_xarray"].attrs["DT"]  # in s
    else:
        dt_from_tracking = 1.0
        logging.warning("Could not find DT in dataset, computing velocity in m/frame")

    # Cell Growth Rate
    cell_growth_info = {"frame": [], "feature_id": [], "cell_id": [], "cell_growth": []}

    # List all of the unique cells and make sure there are no non-physical cells
    cell_groups = Tracks_with_var.groupby("cell_id")

    for cell_id, cell_g in tqdm(
        cell_groups,
        desc="=====Calculating Cell Growth Rates=====",
        total=len(cell_groups),
    ):
        cell_feature_arr = np.array(cell_g["feature_id"]).tolist()
        cell_frame_arr = np.array(cell_g["frame"]).tolist()
        cell_arr = [cell_id] * len(cell_frame_arr)
        cell_var_arr = np.array(cell_g[cell_g.columns[-1]])

        cell_lifetime_arr = np.array(
            cell_g["lifetime"] / np.timedelta64(1, "s")
        )  # convert from ns to s

        # If the cell only lives for one frame, make sure each other array is also one frame, then append nan
        if len(cell_lifetime_arr) == 1:
            assert len(cell_frame_arr) == 1
            assert len(cell_feature_arr) == 1
            assert len(cell_arr) == 1

            cell_growth_info["frame"].append(cell_frame_arr[0])
            cell_growth_info["feature_id"].append(cell_feature_arr[0])
            cell_growth_info["cell_id"].append(cell_arr[0])
            cell_growth_info["cell_growth"].append(np.nan)
            continue

        # If the cell lasts for more than one frame, calculate delta var and delta t
        dvar = cell_var_arr[1:] - cell_var_arr[:-1]
        if type(dt_from_tracking) == np.ndarray or type(dt_from_tracking) == list:

            if type(dt_from_tracking) != np.ndarray:
                dt_from_tracking = np.array(dt_from_tracking)
            # the line below does not work if cell frames are discontinuous
            # dt = dt_from_tracking[cell_frame_arr[0] : cell_frame_arr[-1]] 

            # instead use
            dt = []
            for find, f in enumerate(cell_frame_arr[:-1]):
                dt_across_many_frames = np.sum(dt_from_tracking[f : cell_frame_arr[ find + 1 ]])
                dt.append(dt_across_many_frames)

        else:
            dt = np.diff(cell_frame_arr) * dt_from_tracking

        # The eth/t slope is the growth rate
        slope = dvar / dt
        slope = slope.tolist()
        slope.append(np.nan)

        # Iterate through each element and append them to the info array
        cell_growth_info["frame"].extend(cell_frame_arr)
        cell_growth_info["feature_id"].extend(cell_feature_arr)
        cell_growth_info["cell_id"].extend(cell_arr)
        cell_growth_info["cell_growth"].extend(slope)

    cell_growth_df = pd.DataFrame(cell_growth_info)
    uncut_feature_list = np.unique(cell_growth_df["feature_id"].values)

    # Enforce the threshold
    if threshold is not None:
        # Load background varaible
        if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
            variable_field = analysis_object["segmentation_xarray"]
        else:
            variable_field = analysis_object["segmentation_xarray"][variable]

        if len(variable_field.shape) != 4 and len(variable_field.shape) != 3:
            raise ValueError("=====The segmentation must be 2D or 3D=====")

        feature_mask = analysis_object[f"US_segmentation_{dim}d"].Feature_Segmentation
        feature_mask.values[variable_field.values < threshold] = -1
        uncut_feature_list = np.unique(feature_mask)[1:]  # don"t use -1
        cell_growth_df = cell_growth_df[
            cell_growth_df["feature_id"].isin(uncut_feature_list)
        ]

    cell_growth_df.sort_values("feature_id", inplace=True)
    cell_growth_df.set_index(np.arange(len(cell_growth_df["feature_id"])), inplace=True)

    return cell_growth_df


def calculate_perimeter(  # TODO: split into surface area and perimeter
    analysis_object: dict,
    variable: str | None = None,
    threshold: float | None = None,
    **args: dict,
) -> pd.DataFrame:
    """

    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks.
    variable : str, optional
        Variable to which we should apply the threshold. The default is None.
    threshold : float, optional
        Value of which the area should be greater than. The default is None.
    **args : dict
        Throw away variables.

    Raises
    ------
    Exception
        Exception if missing tracks input from the analysis object.
        Exception if the segmentation data is not 2D or 3D.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, perimeter where perimeter is in km.

    """

    if analysis_object["US_tracks"] is None:
        raise Exception("!=====Tracks Data is Required for Perimeter Calculation=====!")

    # Check if there is 2D segmentation
    if "height" in args and analysis_object["US_segmentation_3d"] is not None:
        altitudes = analysis_object["segmentation_xarray"]["altitudes"]

        height_index = find_nearest(altitudes, args["height"] * 1000)
        feature_seg = analysis_object["US_segmentation_3d"].Feature_Segmentation[
            :, height_index
        ]

        # define some useful variables
        proj_x = analysis_object[f"US_segmentation_3d"].projection_x_coordinate
        proj_y = analysis_object[f"US_segmentation_3d"].projection_y_coordinate

    elif analysis_object["US_segmentation_2d"] is not None:
        feature_seg = analysis_object["US_segmentation_2d"].Feature_Segmentation

        # define some useful variables
        proj_x = analysis_object[f"US_segmentation_2d"].projection_x_coordinate
        proj_y = analysis_object[f"US_segmentation_2d"].projection_y_coordinate

    else:
        raise Exception(
            "!=====Missing Invalid Segmentation Input, needs either 2D segmetation or 3D segmentation with height [km] parameter=====!"
        )

    perimeter_info = {"frame": [], "feature_id": [], "cell_id": [], "perimeter": []}
    Tracks = analysis_object["US_tracks"]

    for ii in tqdm(
        range(len(Tracks)),
        desc="=====Calculating 2D Feature Perimeters=====",
        total=len(Tracks),
    ):
        feature_id = Tracks["feature_id"][ii]
        frame = Tracks["frame"][ii]
        cell_id = Tracks["cell_id"][ii]

        x_dim_sizes = np.diff(proj_x) / 1000  # m->km
        y_dim_sizes = np.diff(proj_y) / 1000  # m->km

        # These are one cell too small due to how diff works, so infer last cell size using the same cell size as the previous cell
        x_dim_sizes = np.append(x_dim_sizes, x_dim_sizes[-1])
        y_dim_sizes = np.append(y_dim_sizes, y_dim_sizes[-1])

        perims = 0  # this is to keep track of the amount of the perimeter that is shared with another cell
        # key is feature id of other cell, value is the shared perimeter

        feature_seg_in_frame = feature_seg[frame].values

        for ny, nx in np.argwhere(feature_seg_in_frame == feature_id):
            for mx in (nx - 1, nx + 1):
                if (
                    mx in range(feature_seg_in_frame.shape[1])
                    and feature_seg_in_frame[ny, mx] != feature_id
                ):
                    perims += y_dim_sizes[ny]

            for my in (ny - 1, ny + 1):
                if (
                    my in range(feature_seg_in_frame.shape[0])
                    and feature_seg_in_frame[my, nx] != feature_id
                ):
                    perims += x_dim_sizes[nx]

        perimeter_info["frame"].append(frame)
        perimeter_info["feature_id"].append(feature_id)
        perimeter_info["cell_id"].append(cell_id)
        perimeter_info["perimeter"].append(perims)

    perimeter_df = pd.DataFrame(perimeter_info)
    # Enforce the threshold
    if threshold is not None:
        # Load background varaible
        if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
            variable_field = analysis_object["segmentation_xarray"]
        else:
            variable_field = analysis_object["segmentation_xarray"][variable]

        if len(variable_field.shape) != 3:
            raise ValueError("=====The segmentation must be 2D=====")

        feature_mask = analysis_object["US_segmentation_2d"].Feature_Segmentation
        # TODO: if user only has 3D segmentation and height for perimeter and has a threshold set
        feature_mask.values[variable_field.values < threshold] = -1
        uncut_feature_list = np.unique(feature_mask)[1:]  # don"t use -1
        perimeter_df = perimeter_df[perimeter_df["feature_id"].isin(uncut_feature_list)]

    return perimeter_df


def calculate_surface_area(
    analysis_object: dict,
    variable: str | None = None,
    threshold: float | None = None,
    **args: dict,
) -> pd.DataFrame:
    """

    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks.
    variable : str, optional
        Variable to which we should apply the threshold. The default is None.
    threshold : float, optional
        Value of which the area should be greater than. The default is None.
    **args : dict
        Throw away variables.

    Raises
    ------
    Exception
        Exception if missing tracks input from the analysis object.
        Exception if the segmentation data is not 2D or 3D.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, surface_area where surface_area is in km^2.

    """

    if analysis_object["US_tracks"] is None:
        raise Exception(
            "!=====Tracks Data is Required for Surface Area Calculation=====!"
        )

    # Check if there is 3D segmentation
    if analysis_object["US_segmentation_3d"] is not None:
        feature_seg = analysis_object["US_segmentation_3d"].Feature_Segmentation

    else:
        raise Exception("!=====Missing 3D Segmentation Input=====!")

    surface_area_info = {
        "frame": [],
        "feature_id": [],
        "cell_id": [],
        "surface_area": [],
    }
    Tracks = analysis_object["US_tracks"]

    # define some useful variables
    proj_x = analysis_object["US_segmentation_3d"].projection_x_coordinate
    proj_y = analysis_object["US_segmentation_3d"].projection_y_coordinate
    altitude = analysis_object["US_segmentation_3d"].altitude

    for ii in tqdm(
        range(len(Tracks)),
        desc="=====Calculating Feature Surface Areas=====",
        total=len(Tracks),
    ):
        feature_id = Tracks["feature_id"][ii]
        frame = Tracks["frame"][ii]
        cell_id = Tracks["cell_id"][ii]

        x_dim_sizes = np.diff(proj_x) / 1000  # m->km
        y_dim_sizes = np.diff(proj_y) / 1000  # m->km
        z_dim_sizes = np.diff(altitude) / 1000  # m->km

        # These are one cell too small due to how diff works, so infer last cell size using the same cell size as the previous cell
        x_dim_sizes = np.append(x_dim_sizes, x_dim_sizes[-1])
        y_dim_sizes = np.append(y_dim_sizes, y_dim_sizes[-1])
        z_dim_sizes = np.append(z_dim_sizes, z_dim_sizes[-1])

        surface_areas = 0  # this is to keep track of the amount of the surface_area that is shared with another cell
        # key is feature id of other cell, value is the shared surface_area

        feature_seg_in_frame = feature_seg[frame].values

        for nz, ny, nx in np.argwhere(feature_seg_in_frame == feature_id):
            for mx in (nx - 1, nx + 1):
                if (
                    mx in range(feature_seg_in_frame.shape[2])
                    and feature_seg_in_frame[nz, ny, mx] != feature_id
                ):
                    surface_areas += y_dim_sizes[ny] * z_dim_sizes[nz]

            for my in (ny - 1, ny + 1):
                if (
                    my in range(feature_seg_in_frame.shape[1])
                    and feature_seg_in_frame[nz, my, nx] != feature_id
                ):
                    surface_areas += x_dim_sizes[nx] * z_dim_sizes[nz]

            for mz in (nz - 1, nz + 1):
                if (
                    mz in range(feature_seg_in_frame.shape[0])
                    and feature_seg_in_frame[mz, ny, nx] != feature_id
                ):
                    surface_areas += x_dim_sizes[nx] * y_dim_sizes[ny]

        surface_area_info["frame"].append(frame)
        surface_area_info["feature_id"].append(feature_id)
        surface_area_info["cell_id"].append(cell_id)
        surface_area_info["surface_area"].append(surface_areas)

    surface_area_df = pd.DataFrame(surface_area_info)

    # Enforce the threshold
    if threshold is not None:
        # Load background varaible
        if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
            variable_field = analysis_object["segmentation_xarray"]
        else:
            variable_field = analysis_object["segmentation_xarray"][variable]

        if len(variable_field.shape) != 4:
            raise ValueError("=====The segmentation must be 3D=====")

        feature_mask = analysis_object[f"US_segmentation_3d"].Feature_Segmentation
        feature_mask.values[variable_field.values < threshold] = -1
        uncut_feature_list = np.unique(feature_mask)[1:]  # don"t use -1
        surface_area_df = surface_area_df[
            surface_area_df["feature_id"].isin(uncut_feature_list)
        ]

    return surface_area_df


def calculate_irregularity(
    analysis_object: dict,
    irregularity_metrics: str | list[str],
    variable: str | None = None,
    threshold: float | None = None,
    segmentation_type: str = None,
    **args: dict,
) -> pd.DataFrame:
    """


    Parameters
    ----------
    analysis_object : dict
        A CoCoMET-US standard analysis object containing at least US_tracks and US_segmentation_2d or US_segmentation_3d.
    irregularity_metrics: str | list[str]
        A string or list of strings for which irregularity metric to calculate.
        The implemented metrics are:
            "sphericity",
            "convexity",
    variable : str, optional
        Variable to which we should apply the threshold. The default is None.
    threshold : float, optional
        Value of which the area should be greater than. The default is None.
    segmentation_type : str, optional
        Whether to calculate 2d or 3d convexity if convexity in irregularity_matrics. If None, will use 3d segmentation if it is available, or 2d segmentation if 3d segmentation is unavailable. The default is None.
    **args : dict
        Throw away variables.

    Raises
    ------
    KeyError
        KeyError if there are no found implemented irregularity metrics.
    Exception
        Exception if no segmentation data is found.

    Returns
    -------
    pandas.core.frame.DataFrame
        A pandas dataframe with the following rows: frame, feature_id, cell_id, irregularity.

    """
    from .irregularity import convexity, sphericity

    if type(irregularity_metrics) != list:
        irregularity_metrics = [irregularity_metrics]

    # A list of the currently implemented irregularity metrics
    implemented_metrics = ["sphericity", "convexity"]

    overlap = set(implemented_metrics) & set(irregularity_metrics)
    if len(overlap) == 0:
        raise KeyError("Found No Implemented Irregularity Metric")

    list_of_irregularities = []

    # Calculate convexity
    if "convexity" in irregularity_metrics:
        if "perimeter" in args:
            perimeter_df = args["perimeter"]["perimeter"]
        else:
            perimeter_df = calculate_perimeter(analysis_object)["perimeter"]

        list_of_irregularities.append(
            convexity.convexity(analysis_object, perimeter_df, segmentation_type)
        )

    # Calculate sphericity
    if "sphericity" in irregularity_metrics:
        # Get volume and perimeter dataframes
        if "volume" in args:
            volume_df = args["volume"]["volume"]
        else:
            volume_df = calculate_volume(analysis_object, None, None, **args)["volume"]

        if "perimeter" in args:
            perimeter_df = args[
                "perimeter"
            ]  # takes the entire perimeter df, not just perimeter values
        else:
            perimeter_df = calculate_perimeter(analysis_object, None, None **args)

        list_of_irregularities.append(sphericity.sphericity(perimeter_df, volume_df))

    # Enforce the threshold
    if analysis_object["US_segmentation_3d"] is not None:
        uncut_feature_list = np.unique(
            analysis_object["US_segmentation_3d"].Feature_Segmentation
        )[
            1:
        ]  # don"t use -1
    elif analysis_object["US_segmentation_2d"] is not None:
        uncut_feature_list = np.unique(
            analysis_object["US_segmentation_2d"].Feature_Segmentation[1:]
        )
    else:
        raise ValueError("!=====Missing US Segmentation Data=====!")

    if threshold is not None:
        # Load background varaible
        if type(analysis_object["segmentation_xarray"]) == xr.core.dataarray.DataArray:
            variable_field = analysis_object["segmentation_xarray"]
        else:
            variable_field = analysis_object["segmentation_xarray"][variable]

        feature_mask = analysis_object["US_segmentation_3d"].Feature_Segmentation
        feature_mask.values[variable_field.values < threshold] = -1
        uncut_feature_list = np.unique(feature_mask)[1:]  # don"t use -1

    irregularity_info_df = list_of_irregularities[0]

    # if there are more irregularity metrics, add them as a column to the irregularity info dataframe
    if len(list_of_irregularities) > 1:
        for df in list_of_irregularities[1:]:
            irregularity_info_df = irregularity_info_df.join(df[df.keys()[-1]])

    irregularity_info_df = irregularity_info_df[
        irregularity_info_df["feature_id"].isin(uncut_feature_list)
    ]

    return irregularity_info_df



import datetime

import geopandas as gpd
import xarray as xr

from .rams_calculate_products import (
    rams_calculate_brightness_temp,
    rams_calculate_precip_rate,
)
from .rams_load import rams_load_netcdf_iris
from .TAMS import run


def rams_run_tams(
    rams_xarray: xr.Dataset, CONFIG: dict
) -> tuple[gpd.GeoDataFrame, tuple]:
    """


    Parameters
    ----------
    rams_xarray : xarray.core.dataset.Dataset
        Xarray Dataset containing RAMS data calculated from rams_load_netcdf(_iris).
    CONFIG : dict
        Default CoCoMET CONFIG file.

    Returns
    -------
    ce : geopandas.geodataframe.GeoDataFrame
        A geopandas dataframe with the identified cloud elements.
    latlon_coord_system : tuple
        A tuple of the latitude and longitude coordinate arrays.

    """

    # make a copy of RAMS dataset to configure it to an acceptable format for TAMS
    rams_for_tams_copy = xr.Dataset({})

    # if brightness temperature is already in rams_xarray use it
    if "TB" not in rams_xarray:
        tb = rams_calculate_brightness_temp(rams_xarray)

    else:
        tb = rams_xarray["TB"]

    rams_for_tams_copy["ctt"] = tb.assign_attrs(
        {"long_name": "Brightness temperature", "units": "K"}
    )
    rams_for_tams_copy["ctt"].chunk(rams_xarray["TOPT"].chunksizes)
    rams_xarray = rams_xarray.assign(TB=rams_for_tams_copy["ctt"])

    # # if precipitation rate is already in rams_xarray use it
    # if "PR" not in rams_xarray:
    #     pr = rams_load_netcdf_iris(CONFIG["rams"]["path_to_data"], "PR", CONFIG["rams"]["path_to_header"], CONFIG)[1]["PR"] # TODO: it is cumbersome to reload the rams_xarray just to calculate precipitation. This needs major reworking

    # else:
    #     pr = rams_xarray["PR"]

    # rams_for_tams_copy["pr"] = pr.assign_attrs(
    #     {"long_name": "Precipitation rate", "units": "mm h-1"}
    # )
    # rams_for_tams_copy["pr"].chunk(rams_xarray["TOPT"].chunksizes)

    # format the times into a list of datetime objects
    dt = rams_xarray.DT
    init_date_str_unformatted = rams_xarray.date[-19:]
    datetime_start_date = datetime.datetime.strptime(
        init_date_str_unformatted, "%Y-%m-%d %H:%M:%S"
    )
    time = []
    for t in rams_xarray.Time.values:
        ds = t * dt
        change = datetime.timedelta(seconds=ds)

        time.append(datetime_start_date + change)

    # make the coordinates and rename the dimensions
    rams_for_tams_copy = rams_for_tams_copy.assign_coords(time=(["Time"], time))
    rams_for_tams_copy = rams_for_tams_copy.assign_coords(
        lat=(["Time", "south_north", "west_east"], rams_xarray["GLAT"].data)
    )
    rams_for_tams_copy = rams_for_tams_copy.assign_coords(
        lon=(["Time", "south_north", "west_east"], rams_xarray["GLON"].data)
    )

    rams_for_tams_copy = rams_for_tams_copy.rename(
        {"Time": "time", "south_north": "lat", "west_east": "lon"}
    )  # rename the dimensions

    ce, latlon_coord_system = run(rams_for_tams_copy, **CONFIG["rams"]["tams"])

    return (ce, latlon_coord_system)



# =============================================================================
# This defines the methods for running MOAAP on RAMS data processed using rams_load.py
# =============================================================================

import os

import numpy as np
import pandas as pd
import xarray as xr

from CoCoMET.MOAAP import moaap

from .rams_calculate_products import (
    rams_calculate_brightness_temp,
    rams_calculate_precip_rate,
    rams_calculate_reflectivity,
    rams_calculate_wa,
)


# Calculate nearest item in list to given pivot
def find_nearest(array, pivot):
    import numpy as np

    array = np.asarray(array)
    idx = (np.abs(array - pivot)).argmin()
    return idx


def rams_run_moaap(rams_xarray, CONFIG):
    """
    Inputs:
        rams_xarray: xarray Dataset containing RAMS data calculated from rams_load.py
        CONFIG: User configuration file
    Outputs:
        mask_file: The xarray object containing the default MOAAP outputs
    """

    # Get basic setup variables including lat/lon, delta time, a pandas time range vector (TODO: adjust output to )
    latitudes = rams_xarray.GLAT[0].values
    longitudes = rams_xarray.GLON[0].values
    dt = rams_xarray.DT
    times = pd.date_range(
        start=rams_xarray["Time"].values[0],
        end=rams_xarray["Time"].values[-1],
        freq=str(dt) + "min",
    )
    mask = np.ones(latitudes.shape)

    # Get all necessary variables from RAMS output to input into MOAAP

    # Get pressure heights
    PI = rams_xarray["PI"]
    p00 = 1e5  # Pa
    cp = 1004  # J / kg / K
    Rd = 287  # J / kg / K

    total_p = (PI / cp) ** (Rd / cp) * p00 / 100
    avg_geo_pres = np.mean(total_p[0, :, :, :].values, axis=(1, 2))

    # Get height_idx of 850hPA, 500hPA, and 200hPA
    height_idx_850 = find_nearest(avg_geo_pres, 850)
    height_idx_500 = find_nearest(avg_geo_pres, 500)
    height_idx_200 = find_nearest(avg_geo_pres, 200)

    # get destaggered 850hPA wind speeds and 200hPA wind speeds
    v_winds_850 = (rams_xarray["VC"][:, height_idx_850, :, :]).values
    u_winds_850 = (rams_xarray["UC"][:, height_idx_850, :, :]).values
    v_winds_200 = (rams_xarray["VC"][:, height_idx_200, :, :]).values
    u_winds_200 = (rams_xarray["UC"][:, height_idx_200, :, :]).values

    # Calculate proper pressures and actual temperature
    full_t = rams_xarray["THETA"]

    air_temp = (full_t * (total_p / 1e5) ** (287.0 / 1004.5))[:, height_idx_850].values

    # Get geopt
    geopt = rams_xarray["geopt"]

    # Get brightness temp
    tb = (
        rams_xarray.TB.values
        if "TB" in rams_xarray
        else rams_calculate_brightness_temp(rams_xarray)
    )

    # Get total mixing ratio at 850hPA
    mr = (
        rams_xarray.RDP
        + rams_xarray.RCP
        + rams_xarray.RRP
        + rams_xarray.RPP
        + rams_xarray.RSP
        + rams_xarray.RGP
        + rams_xarray.RHP
        + rams_xarray.RAP
    )[
        :, height_idx_850
    ] * 1000  # kg / kg -> g / kg

    # Get precipitation rate
    pr = rams_calculate_precip_rate(rams_xarray) / (60 / rams_xarray.DT)

    # Create output folder if it does not already exist
    if not os.path.exists(CONFIG["rams"]["moaap"]["tracking_save_path"]):
        os.makedirs(CONFIG["rams"]["moaap"]["tracking_save_path"])

    moaap(
        longitudes,
        latitudes,
        times,
        dt / 3600,  # I think it's supposed to be in hours
        mask,
        DataName="CoMET_RAMS_MOAAP_TRACKING",
        OutputFolder=CONFIG["rams"]["moaap"]["tracking_save_path"],
        # Data Variables
        v850=v_winds_850,
        u850=u_winds_850,
        t850=air_temp,
        q850=mr,
        slp=total_p[:, 0, :, :],
        ivte=None,
        ivtn=None,
        z500=geopt.values[:, height_idx_500],
        v200=v_winds_200,
        u200=u_winds_200,
        pr=pr,
        tb=tb,
        # Any user defined params
        **CONFIG["rams"]["moaap"],
    )

    if CONFIG["rams"]["moaap"]["tracking_save_path"][-1] != "/":
        CONFIG["rams"]["moaap"]["tracking_save_path"] += "/"

    output_filepath = (
        CONFIG["rams"]["moaap"]["tracking_save_path"]
        + str(times[0].year)
        + str(times[0].month).zfill(2)
        + "_CoMET_RAMS_MOAAP_TRACKING_ObjectMasks_dt-%.2f" % dt
        + "min_MOAAP-masks"
        + ".nc"
    )

    mask_file = xr.open_mfdataset(
        output_filepath,
        coords="all",
        concat_dim="time",
        combine="nested",
        decode_time=False,
    )

    return mask_file
